"""Flic client library for python

Requires python 3.3 or higher.

For detailed documentation, see the protocol documentation.

Notes on the data type used in this python implementation compared to the protocol documentation:
All kind of integers are represented as python integers.
Booleans use the Boolean type.
Enums use the defined python enums below.
Bd addr are represented as standard python strings, e.g. "aa:bb:cc:dd:ee:ff".
"""

from enum import Enum
from collections import namedtuple
import time
import socket
import select
import struct
import itertools
import queue
import threading

class CreateConnectionChannelError(Enum):
	NoError = 0
	MaxPendingConnectionsReached = 1

class ConnectionStatus(Enum):
	Disconnected = 0
	Connected = 1
	Ready = 2

class DisconnectReason(Enum):
	Unspecified = 0
	ConnectionEstablishmentFailed = 1
	TimedOut = 2
	BondingKeysMismatch = 3

class RemovedReason(Enum):
	RemovedByThisClient = 0
	ForceDisconnectedByThisClient = 1
	ForceDisconnectedByOtherClient = 2
	
	ButtonIsPrivate = 3
	VerifyTimeout = 4
	InternetBackendError = 5
	InvalidData = 6
	
	CouldntLoadDevice = 7

class ClickType(Enum):
	ButtonDown = 0
	ButtonUp = 1
	ButtonClick = 2
	ButtonSingleClick = 3
	ButtonDoubleClick = 4
	ButtonHold = 5

class BdAddrType(Enum):
	PublicBdAddrType = 0
	RandomBdAddrType = 1

class LatencyMode(Enum):
	NormalLatency = 0
	LowLatency = 1
	HighLatency = 2

class BluetoothControllerState(Enum):
	Detached = 0
	Resetting = 1
	Attached = 2

class ScanWizardResult(Enum):
	WizardSuccess = 0
	WizardCancelledByUser = 1
	WizardFailedTimeout = 2
	WizardButtonIsPrivate = 3
	WizardBluetoothUnavailable = 4
	WizardInternetBackendError = 5
	WizardInvalidData = 6

class ButtonScanner:
	"""ButtonScanner class.
	
	Usage:
	scanner = ButtonScanner()
	scanner.on_advertisement_packet = lambda scanner, bd_addr, name, rssi, is_private, already_verified: ...
	client.add_scanner(scanner)
	"""
	
	_cnt = itertools.count()
	
	def __init__(self):
		self._scan_id = next(ButtonScanner._cnt)
		self.on_advertisement_packet = lambda scanner, bd_addr, name, rssi, is_private, already_verified: None

class ScanWizard:
	"""ScanWizard class
	
	Usage:
	wizard = ScanWizard()
	wizard.on_found_private_button = lambda scan_wizard: ...
	wizard.on_found_public_button = lambda scan_wizard, bd_addr, name: ...
	wizard.on_button_connected = lambda scan_wizard, bd_addr, name: ...
	wizard.on_completed = lambda scan_wizard, result, bd_addr, name: ...
	client.add_scan_wizard(wizard)
	"""
	
	_cnt = itertools.count()
	
	def __init__(self):
		self._scan_wizard_id = next(ScanWizard._cnt)
		self._bd_addr = None
		self._name = None
		self.on_found_private_button = lambda scan_wizard: None
		self.on_found_public_button = lambda scan_wizard, bd_addr, name: None
		self.on_button_connected = lambda scan_wizard, bd_addr, name: None
		self.on_completed = lambda scan_wizard, result, bd_addr, name: None

class ButtonConnectionChannel:
	"""ButtonConnectionChannel class.
	
	This class represents a connection channel to a Flic button.
	Add this button connection channel to a FlicClient by executing client.add_connection_channel(connection_channel).
	You may only have this connection channel added to one FlicClient at a time.
	
	Before you add the connection channel to the client, you should set up your callback functions by assigning
	the corresponding properties to this object with a function. Each callback function has a channel parameter as the first one,
	referencing this object.
	
	Available properties and the function parameters are:
	on_create_connection_channel_response: channel, error, connection_status
	on_removed: channel, removed_reason
	on_connection_status_changed: channel, connection_status, disconnect_reason
	on_button_up_or_down / on_button_click_or_hold / on_button_single_or_double_click / on_button_single_or_double_click_or_hold: channel, click_type, was_queued, time_diff
	"""
	
	_cnt = itertools.count()
	
	def __init__(self, bd_addr, latency_mode = LatencyMode.NormalLatency, auto_disconnect_time = 511):
		self._conn_id = next(ButtonConnectionChannel._cnt)
		self._bd_addr = bd_addr
		self._latency_mode = latency_mode
		self._auto_disconnect_time = auto_disconnect_time
		self._client = None
		
		self.on_create_connection_channel_response = lambda channel, error, connection_status: None
		self.on_removed = lambda channel, removed_reason: None
		self.on_connection_status_changed = lambda channel, connection_status, disconnect_reason: None
		self.on_button_up_or_down = lambda channel, click_type, was_queued, time_diff: None
		self.on_button_click_or_hold = lambda channel, click_type, was_queued, time_diff: None
		self.on_button_single_or_double_click = lambda channel, click_type, was_queued, time_diff: None
		self.on_button_single_or_double_click_or_hold = lambda channel, click_type, was_queued, time_diff: None
	
	@property
	def bd_addr(self):
		return self._bd_addr
	
	@property
	def latency_mode(self):
		return self._latency_mode
	
	@latency_mode.setter
	def latency_mode(self, latency_mode):
		if self._client is None:
			self._latency_mode = latency_mode
			return
		
		with self._client._lock:
			self._latency_mode = latency_mode
			if not self._client._closed:
				self._client._send_command("CmdChangeModeParameters", {"conn_id": self._conn_id, "latency_mode": self._latency_mode, "auto_disconnect_time": self._auto_disconnect_time})
	
	@property
	def auto_disconnect_time(self):
		return self._auto_disconnect_time
	
	@auto_disconnect_time.setter
	def auto_disconnect_time(self, auto_disconnect_time):
		if self._client is None:
			self._auto_disconnect_time = auto_disconnect_time
			return
		
		with self._client._lock:
			self._auto_disconnect_time = auto_disconnect_time
			if not self._client._closed:
				self._client._send_command("CmdChangeModeParameters", {"conn_id": self._conn_id, "latency_mode": self._latency_mode, "auto_disconnect_time": self._auto_disconnect_time})

class FlicClient:
	"""FlicClient class.
	
	When this class is constructed, a socket connection is established.
	You may then send commands to the server and set timers.
	Once you are ready with the initialization you must call the handle_events() method which is a main loop that never exits, unless the socket is closed.
	For a more detailed description of all commands, events and enums, check the protocol specification.
	
	All commands are wrapped in more high level functions and events are reported using callback functions.
	
	All methods called on this class will take effect only if you eventually call the handle_events() method.
	
	The ButtonScanner is used to set up a handler for advertisement packets.
	The ButtonConnectionChannel is used to interact with connections to flic buttons and receive their events.
	
	Other events are handled by the following callback functions that can be assigned to this object (and a list of the callback function parameters):
	on_new_verified_button: bd_addr
	on_no_space_for_new_connection: max_concurrently_connected_buttons
	on_got_space_for_new_connection: max_concurrently_connected_buttons
	on_bluetooth_controller_state_change: state
	"""
	
	_EVENTS = [
		("EvtAdvertisementPacket", "<I6s17pb??", "scan_id bd_addr name rssi is_private already_verified"),
		("EvtCreateConnectionChannelResponse", "<IBB", "conn_id error connection_status"),
		("EvtConnectionStatusChanged", "<IBB", "conn_id connection_status disconnect_reason"),
		("EvtConnectionChannelRemoved", "<IB", "conn_id removed_reason"),
		("EvtButtonUpOrDown", "<IBBI", "conn_id click_type was_queued time_diff"),
		("EvtButtonClickOrHold", "<IBBI", "conn_id click_type was_queued time_diff"),
		("EvtButtonSingleOrDoubleClick", "<IBBI", "conn_id click_type was_queued time_diff"),
		("EvtButtonSingleOrDoubleClickOrHold", "<IBBI", "conn_id click_type was_queued time_diff"),
		("EvtNewVerifiedButton", "<6s", "bd_addr"),
		("EvtGetInfoResponse", "<B6sBBhBBH", "bluetooth_controller_state my_bd_addr my_bd_addr_type max_pending_connections max_concurrently_connected_buttons current_pending_connections currently_no_space_for_new_connection nb_verified_buttons"),
		("EvtNoSpaceForNewConnection", "<B", "max_concurrently_connected_buttons"),
		("EvtGotSpaceForNewConnection", "<B", "max_concurrently_connected_buttons"),
		("EvtBluetoothControllerStateChange", "<B", "state"),
		("EvtPingResponse", "<I", "ping_id"),
		("EvtGetButtonUUIDResponse", "<6s16s", "bd_addr uuid"),
		("EvtScanWizardFoundPrivateButton", "<I", "scan_wizard_id"),
		("EvtScanWizardFoundPublicButton", "<I6s17p", "scan_wizard_id bd_addr name"),
		("EvtScanWizardButtonConnected", "<I", "scan_wizard_id"),
		("EvtScanWizardCompleted", "<IB", "scan_wizard_id result")
	]
	_EVENT_STRUCTS = list(map(lambda x: None if x == None else struct.Struct(x[1]), _EVENTS))
	_EVENT_NAMED_TUPLES = list(map(lambda x: None if x == None else namedtuple(x[0], x[2]), _EVENTS))
	
	_COMMANDS = [
		("CmdGetInfo", "", ""),
		("CmdCreateScanner", "<I", "scan_id"),
		("CmdRemoveScanner", "<I", "scan_id"),
		("CmdCreateConnectionChannel", "<I6sBh", "conn_id bd_addr latency_mode auto_disconnect_time"),
		("CmdRemoveConnectionChannel", "<I", "conn_id"),
		("CmdForceDisconnect", "<6s", "bd_addr"),
		("CmdChangeModeParameters", "<IBh", "conn_id latency_mode auto_disconnect_time"),
		("CmdPing", "<I", "ping_id"),
		("CmdGetButtonUUID", "<6s", "bd_addr"),
		("CmdCreateScanWizard", "<I", "scan_wizard_id"),
		("CmdCancelScanWizard", "<I", "scan_wizard_id")
	]
	
	_COMMAND_STRUCTS = list(map(lambda x: struct.Struct(x[1]), _COMMANDS))
	_COMMAND_NAMED_TUPLES = list(map(lambda x: namedtuple(x[0], x[2]), _COMMANDS))
	_COMMAND_NAME_TO_OPCODE = dict((x[0], i) for i, x in enumerate(_COMMANDS))
	
	def _bdaddr_bytes_to_string(bdaddr_bytes):
		return ":".join(map(lambda x: "%02x" % x, reversed(bdaddr_bytes)))
	
	def _bdaddr_string_to_bytes(bdaddr_string):
		return bytearray.fromhex("".join(reversed(bdaddr_string.split(":"))))
	
	def __init__(self, host, port = 5551):
		self._sock = socket.create_connection((host, port), None)
		self._lock = threading.RLock()
		self._scanners = {}
		self._scan_wizards = {}
		self._connection_channels = {}
		self._get_info_response_queue = queue.Queue()
		self._get_button_uuid_queue = queue.Queue()
		self._timers = queue.PriorityQueue()
		self._handle_event_thread_ident = None
		self._closed = False
		
		self.on_new_verified_button = lambda bd_addr: None
		self.on_no_space_for_new_connection = lambda max_concurrently_connected_buttons: None
		self.on_got_space_for_new_connection = lambda max_concurrently_connected_buttons: None
		self.on_bluetooth_controller_state_change = lambda state: None
	
	def close(self):
		"""Closes the client. The handle_events() method will return."""
		with self._lock:
			if self._closed:
				return
			
			if threading.get_ident() != self._handle_event_thread_ident:
				self._send_command("CmdPing", {"ping_id": 0}) # To unblock socket select
			
			self._closed = True
	
	def add_scanner(self, scanner):
		"""Add a ButtonScanner object.
		
		The scan will start directly once the scanner is added.
		"""
		with self._lock:
			if scanner._scan_id in self._scanners:
				return
			
			self._scanners[scanner._scan_id] = scanner
			self._send_command("CmdCreateScanner", {"scan_id": scanner._scan_id})
	
	def remove_scanner(self, scanner):
		"""Remove a ButtonScanner object.
		
		You will no longer receive advertisement packets.
		"""
		with self._lock:
			if scanner._scan_id not in self._scanners:
				return
			
			del self._scanners[scanner._scan_id]
			self._send_command("CmdRemoveScanner", {"scan_id": scanner._scan_id})
	
	def add_scan_wizard(self, scan_wizard):
		"""Add a ScanWizard object.
		
		The scan wizard will start directly once the scan wizard is added.
		"""
		with self._lock:
			if scan_wizard._scan_wizard_id in self._scan_wizards:
				return
			
			self._scan_wizards[scan_wizard._scan_wizard_id] = scan_wizard
			self._send_command("CmdCreateScanWizard", {"scan_wizard_id": scan_wizard._scan_wizard_id})
	
	def cancel_scan_wizard(self, scan_wizard):
		"""Cancel a ScanWizard.
		
		Note: The effect of this command will take place at the time the on_completed event arrives on the scan wizard object.
		If cancelled due to this command, "result" in the on_completed event will be "WizardCancelledByUser".
		"""
		with self._lock:
			if scan_wizard._scan_wizard_id not in self._scan_wizards:
				return
			
			self._send_command("CmdCancelScanWizard", {"scan_wizard_id": scan_wizard._scan_wizard_id})
	
	def add_connection_channel(self, channel):
		"""Adds a connection channel to a specific Flic button.
		
		This will start listening for a specific Flic button's connection and button events.
		Make sure the Flic is either in public mode (by holding it down for 7 seconds) or already verified before calling this method.
		
		The on_create_connection_channel_response callback property will be called on the
		connection channel after this command has been received by the server.
		
		You may have as many connection channels as you wish for a specific Flic Button.
		"""
		with self._lock:
			if channel._conn_id in self._connection_channels:
				return
			
			channel._client = self
			
			self._connection_channels[channel._conn_id] = channel
			self._send_command("CmdCreateConnectionChannel", {"conn_id": channel._conn_id, "bd_addr": channel.bd_addr, "latency_mode": channel._latency_mode, "auto_disconnect_time": channel._auto_disconnect_time})
	
	def remove_connection_channel(self, channel):
		"""Remove a connection channel.
		
		This will stop listening for new events for a specific connection channel that has previously been added.
		Note: The effect of this command will take place at the time the on_removed event arrives on the connection channel object.
		"""
		with self._lock:
			if channel._conn_id not in self._connection_channels:
				return
			
			self._send_command("CmdRemoveConnectionChannel", {"conn_id": channel._conn_id})
	
	def force_disconnect(self, bd_addr):
		"""Force disconnection or cancel pending connection of a specific Flic button.
		
		This removes all connection channels for all clients connected to the server for this specific Flic button.
		"""
		self._send_command("CmdForceDisconnect", {"bd_addr": bd_addr})
	
	def get_info(self, callback):
		"""Get info about the current state of the server.
		
		The server will send back its information directly and the callback will be called once the response arrives.
		The callback takes only one parameter: info. This info parameter is a dictionary with the following objects:
		bluetooth_controller_state, my_bd_addr, my_bd_addr_type, max_pending_connections, max_concurrently_connected_buttons,
		current_pending_connections, currently_no_space_for_new_connection, bd_addr_of_verified_buttons (a list of bd addresses).
		"""
		self._get_info_response_queue.put(callback)
		self._send_command("CmdGetInfo", {})
	
	def get_button_uuid(self, bd_addr, callback):
		"""Get button uuid for a verified button.
		
		The server will send back its information directly and the callback will be called once the response arrives.
		Responses will arrive in the same order as requested.
		
		The callback takes two parameters: bd_addr, uuid (hex string of 32 characters).
		
		Note: if the button isn't verified, the uuid sent to the callback will rather be None.
		"""
		with self._lock:
			self._get_button_uuid_queue.put(callback)
			self._send_command("CmdGetButtonUUID", {"bd_addr": bd_addr})
	
	def set_timer(self, timeout_millis, callback):
		"""Set a timer
		
		This timer callback will run after the specified timeout_millis on the thread that handles the events.
		"""
		point_in_time = time.monotonic() + timeout_millis / 1000.0
		self._timers.put((point_in_time, callback))
		
		if threading.get_ident() != self._handle_event_thread_ident:
			self._send_command("CmdPing", {"ping_id": 0}) # To unblock socket select
	
	def run_on_handle_events_thread(self, callback):
		"""Run a function on the thread that handles the events."""
		if threading.get_ident() == self._handle_event_thread_ident:
			callback()
		else:
			self.set_timer(0, callback)
	
	def _send_command(self, name, items):
		for key, value in items.items():
			if isinstance(value, Enum):
				items[key] = value.value
		
		if "bd_addr" in items:
			items["bd_addr"] = FlicClient._bdaddr_string_to_bytes(items["bd_addr"])
		
		opcode = FlicClient._COMMAND_NAME_TO_OPCODE[name]
		data_bytes = FlicClient._COMMAND_STRUCTS[opcode].pack(*FlicClient._COMMAND_NAMED_TUPLES[opcode](**items))
		bytes = bytearray(3)
		bytes[0] = (len(data_bytes) + 1) & 0xff
		bytes[1] = (len(data_bytes) + 1) >> 8
		bytes[2] = opcode
		bytes += data_bytes
		with self._lock:
			if not self._closed:
				self._sock.sendall(bytes)
	
	def _dispatch_event(self, data):
		if len(data) == 0:
			return
		opcode = data[0]
		
		if opcode >= len(FlicClient._EVENTS) or FlicClient._EVENTS[opcode] == None:
			return
		
		event_name = FlicClient._EVENTS[opcode][0]
		data_tuple = FlicClient._EVENT_STRUCTS[opcode].unpack(data[1 : 1 + FlicClient._EVENT_STRUCTS[opcode].size])
		items = FlicClient._EVENT_NAMED_TUPLES[opcode]._make(data_tuple)._asdict()
		
		# Process some kind of items whose data type is not supported by struct
		if "bd_addr" in items:
			items["bd_addr"] = FlicClient._bdaddr_bytes_to_string(items["bd_addr"])
		
		if "name" in items:
			items["name"] = items["name"].decode("utf-8")
		
		if event_name == "EvtCreateConnectionChannelResponse":
			items["error"] = CreateConnectionChannelError(items["error"])
			items["connection_status"] = ConnectionStatus(items["connection_status"])
		
		if event_name == "EvtConnectionStatusChanged":
			items["connection_status"] = ConnectionStatus(items["connection_status"])
			items["disconnect_reason"] = DisconnectReason(items["disconnect_reason"])
		
		if event_name == "EvtConnectionChannelRemoved":
			items["removed_reason"] = RemovedReason(items["removed_reason"])
		
		if event_name.startswith("EvtButton"):
			items["click_type"] = ClickType(items["click_type"])
		
		if event_name == "EvtGetInfoResponse":
			items["bluetooth_controller_state"] = BluetoothControllerState(items["bluetooth_controller_state"])
			items["my_bd_addr"] = FlicClient._bdaddr_bytes_to_string(items["my_bd_addr"])
			items["my_bd_addr_type"] = BdAddrType(items["my_bd_addr_type"])
			items["bd_addr_of_verified_buttons"] = []
			
			pos = FlicClient._EVENT_STRUCTS[opcode].size
			for i in range(items["nb_verified_buttons"]):
				items["bd_addr_of_verified_buttons"].append(FlicClient._bdaddr_bytes_to_string(data[1 + pos : 1 + pos + 6]))
				pos += 6
		
		if event_name == "EvtBluetoothControllerStateChange":
			items["state"] = BluetoothControllerState(items["state"])
		
		if event_name == "EvtGetButtonUUIDResponse":
			items["uuid"] = "".join(map(lambda x: "%02x" % x, items["uuid"]))
			if items["uuid"] == "00000000000000000000000000000000":
				items["uuid"] = None
		
		if event_name == "EvtScanWizardCompleted":
			items["result"] = ScanWizardResult(items["result"])
		
		# Process event
		if event_name == "EvtAdvertisementPacket":
			scanner = self._scanners.get(items["scan_id"])
			if scanner is not None:
				scanner.on_advertisement_packet(scanner, items["bd_addr"], items["name"], items["rssi"], items["is_private"], items["already_verified"])
		
		if event_name == "EvtCreateConnectionChannelResponse":
			channel = self._connection_channels[items["conn_id"]]
			if items["error"] != CreateConnectionChannelError.NoError:
				del self._connection_channels[items["conn_id"]]
			channel.on_create_connection_channel_response(channel, items["error"], items["connection_status"])
		
		if event_name == "EvtConnectionStatusChanged":
			channel = self._connection_channels[items["conn_id"]]
			channel.on_connection_status_changed(channel, items["connection_status"], items["disconnect_reason"])
		
		if event_name == "EvtConnectionChannelRemoved":
			channel = self._connection_channels[items["conn_id"]]
			del self._connection_channels[items["conn_id"]]
			channel.on_removed(channel, items["removed_reason"])
		
		if event_name == "EvtButtonUpOrDown":
			channel = self._connection_channels[items["conn_id"]]
			channel.on_button_up_or_down(channel, items["click_type"], items["was_queued"], items["time_diff"])
		if event_name == "EvtButtonClickOrHold":
			channel = self._connection_channels[items["conn_id"]]
			channel.on_button_click_or_hold(channel, items["click_type"], items["was_queued"], items["time_diff"])
		if event_name == "EvtButtonSingleOrDoubleClick":
			channel = self._connection_channels[items["conn_id"]]
			channel.on_button_single_or_double_click(channel, items["click_type"], items["was_queued"], items["time_diff"])
		if event_name == "EvtButtonSingleOrDoubleClickOrHold":
			channel = self._connection_channels[items["conn_id"]]
			channel.on_button_single_or_double_click_or_hold(channel, items["click_type"], items["was_queued"], items["time_diff"])
		
		if event_name == "EvtNewVerifiedButton":
			self.on_new_verified_button(items["bd_addr"])
		
		if event_name == "EvtGetInfoResponse":
			self._get_info_response_queue.get()(items)
		
		if event_name == "EvtNoSpaceForNewConnection":
			self.on_no_space_for_new_connection(items["max_concurrently_connected_buttons"])
		
		if event_name == "EvtGotSpaceForNewConnection":
			self.on_got_space_for_new_connection(items["max_concurrently_connected_buttons"])
		
		if event_name == "EvtBluetoothControllerStateChange":
			self.on_bluetooth_controller_state_change(items["state"])
		
		if event_name == "EvtGetButtonUUIDResponse":
			self._get_button_uuid_queue.get()(items["bd_addr"], items["uuid"])
		
		if event_name == "EvtScanWizardFoundPrivateButton":
			scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
			scan_wizard.on_found_private_button(scan_wizard)
		
		if event_name == "EvtScanWizardFoundPublicButton":
			scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
			scan_wizard._bd_addr = items["bd_addr"]
			scan_wizard._name = items["name"]
			scan_wizard.on_found_public_button(scan_wizard, scan_wizard._bd_addr, scan_wizard._name)
		
		if event_name == "EvtScanWizardButtonConnected":
			scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
			scan_wizard.on_button_connected(scan_wizard, scan_wizard._bd_addr, scan_wizard._name)
		
		if event_name == "EvtScanWizardCompleted":
			scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
			del self._scan_wizards[items["scan_wizard_id"]]
			scan_wizard.on_completed(scan_wizard, items["result"], scan_wizard._bd_addr, scan_wizard._name)
	
	def _handle_one_event(self):
		if len(self._timers.queue) > 0:
			current_timer = self._timers.queue[0]
			timeout = max(current_timer[0] - time.monotonic(), 0)
			if timeout == 0:
				self._timers.get()[1]()
				return True
			if len(select.select([self._sock], [], [], timeout)[0]) == 0:
				return True
		
		len_arr = bytearray(2)
		view = memoryview(len_arr)
		
		toread = 2
		while toread > 0:
			nbytes = self._sock.recv_into(view, toread)
			if nbytes == 0:
				return False
			view = view[nbytes:]
			toread -= nbytes
		
		packet_len = len_arr[0] | (len_arr[1] << 8)
		data = bytearray(packet_len)
		view = memoryview(data)
		toread = packet_len
		while toread > 0:
			nbytes = self._sock.recv_into(view, toread)
			if nbytes == 0:
				return False
			view = view[nbytes:]
			toread -= nbytes
		
		self._dispatch_event(data)
		return True
		
	def handle_events(self):
		"""Start the main loop for this client.
		
		This method will not return until the socket has been closed.
		Once it has returned, any use of this FlicClient is illegal.
		"""
		self._handle_event_thread_ident = threading.get_ident()
		while not self._closed:
			if not self._handle_one_event():
				break
		self._sock.close()



import os

# noinspection PyPackageRequirements
from PyOBEX.server import PushServer

from platypush.backend.bluetooth import BluetoothBackend


class BluetoothPushserverBackend(BluetoothBackend, PushServer):
    """
    Bluetooth OBEX push server.
    Enable it to allow bluetooth file transfers from other devices.

    If you run platypush as a non-root user (and you should) then you to change the group owner of the
    service discovery protocol file (/var/run/sdp) and add your user to that group. See
    `here <https://stackoverflow.com/questions/34599703/rfcomm-bluetooth-permission-denied-error-raspberry-pi>`_
    for details.

    Requires:

        * **pybluez** (``pip install pybluez``)
        * **pyobex** (``pip install git+https://github.com/BlackLight/PyOBEX``)

    """

    _sleep_on_error = 10.0

    def __init__(self, port: int, address: str = '',
                 directory: str = os.path.join(os.path.expanduser('~'), 'bluetooth'),
                 whitelisted_addresses: list = None, **kwargs):
        """
        :param port: Bluetooth listen port
        :param address: Bluetooth address to bind the server to (default: any)
        :param directory: Destination directory where files will be downloaded (default: ~/bluetooth)
        :param whitelisted_addresses: If set then only accept connections from the listed device addresses
        """
        BluetoothBackend.__init__(self, address=address, port=port, directory=directory,
                                  whitelisted_addresses=whitelisted_addresses, **kwargs)

    def run(self):
        if not os.path.isdir(self.directory):
            os.makedirs(self.directory, exist_ok=True)

        super().run()


# vim:sw=4:ts=4:et:



import enum
import threading
import time

from typing import Dict, List

from platypush.context import get_bus
from platypush.message.event.zeroborg import ZeroborgDriveEvent, ZeroborgStopEvent
from platypush.plugins import Plugin, action


class Direction(enum.Enum):
    DIR_UP = 'up'
    DIR_DOWN = 'down'
    DIR_LEFT = 'left'
    DIR_RIGHT = 'right'


# noinspection PyPep8Naming
class GpioZeroborgPlugin(Plugin):
    """
    ZeroBorg plugin. It allows you to control a ZeroBorg
    (https://www.piborg.org/motor-control-1135/zeroborg) motor controller and
    infrared sensor circuitry for Raspberry Pi

    Triggers:

        * :class:`platypush.message.event.zeroborg.ZeroborgDriveEvent` when motors direction changes
        * :class:`platypush.message.event.zeroborg.ZeroborgStopEvent` upon motors stop

    """

    def __init__(self, directions: Dict[str, List[float]] = None, **kwargs):
        """
        :param directions: Configuration for the motor directions. A direction is basically a configuration of the
            power delivered to each motor to allow whichever object you're controlling (wheels, robotic arms etc.) to
            move in a certain direction. In my experience the ZeroBorg always needs a bit of calibration, depending on
            factory defaults and the mechanical properties of the load it controls.

        Example configuration that I use to control a simple 4WD robot::

            directions = {
                "up": [
                    0.4821428571428572,   # Motor 1 power
                    0.4821428571428572,   # Motor 2 power
                    -0.6707142857142858,  # Motor 3 power
                    -0.6707142857142858   # Motor 4 power
                ],
                "down": [
                    -0.4821428571428572,
                    -0.4821428571428572,
                    0.825,
                    0.825
                ],
                "left": [
                    -0.1392857142857143,
                    -0.1392857142857143,
                    -1.0553571428571429,
                    -1.0553571428571429
                ],
                "right": [
                    1.0017857142857143,
                    1.0017857142857143,
                    0.6214285714285713,
                    0.6214285714285713
                ]
            }

        """

        if directions is None:
            directions = {}

        import platypush.plugins.gpio.zeroborg.lib as ZeroBorg
        super().__init__(**kwargs)

        self.directions = directions
        self._direction = None
        self._drive_thread = None
        self._motors = [0, 0, 0, 0]

        self.zb = ZeroBorg.ZeroBorg()
        self.zb.Init()
        self.zb.SetCommsFailsafe(True)
        self.zb.ResetEpo()

    @staticmethod
    def _get_measurement(plugin, timeout):
        measure_start_time = time.time()
        value = None

        while value is None:
            value = plugin.get_measurement().output
            if time.time() - measure_start_time > timeout:
                return None

        return value

    @action
    def drive(self, direction):
        """
        Drive the motors in a certain direction.
        """

        def _run():
            try:
                while self._direction:
                    try:
                        if self._direction in self.directions:
                            self._motors = self.directions[self._direction]
                        else:
                            self.logger.warning('Invalid direction {}: stopping motors'.format(self._direction))
                    except Exception as e:
                        self.logger.error('Error on _get_direction_from_sensors: {}'.format(str(e)))
                        break

                    for i, power in enumerate(self._motors):
                        method = getattr(self.zb, 'SetMotor{}'.format(i+1))
                        method(power)
            finally:
                self.zb.MotorsOff()
                self.zb.ResetEpo()
                self._drive_thread = None

        self._direction = direction.lower()

        if not self._drive_thread:
            drive_thread = threading.Thread(target=_run)
            drive_thread.start()
            self._drive_thread = drive_thread

        get_bus().post(ZeroborgDriveEvent(direction=self._direction, motors=self.directions[self._direction]))
        return {'status': 'running', 'direction': direction}

    @action
    def stop(self):
        """
        Turns off the motors
        """

        self._direction = None
        if self._drive_thread:
            self._drive_thread.join()

        get_bus().post(ZeroborgStopEvent())
        return {'status': 'stopped'}

    @action
    def status(self) -> dict:
        """
        Get the current direction and motors power. Example response::

            .. code-block:: json

            {
                "status": "running",
                "direction": "up",
                "motors": [1.0, 1.0, -1.0, -1.0]
            }

        """

        return {
            'status': 'running' if self._direction else 'stopped',
            'direction': self._direction,
            'motors': [getattr(self.zb, 'GetMotor{}'.format(i+1))() for i in range(4)],
        }


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorLtr559Backend(SensorBackend):
    """
    Backend to poll an `LTR559 <https://shop.pimoroni.com/products/ltr-559-light-proximity-sensor-breakout>`_
    light/proximity sensor

    Requires:

        * ``ltr559`` (``pip install ltr559``)
    """

    def __init__(self, light=True, proximity=True, **kwargs):
        """
        :param light: Enable light sensor
        :param proximity: Enable proximity sensor
        """

        enabled_sensors = {
            'light': light,
            'proximity': proximity,
        }

        super().__init__(plugin='gpio.sensor.ltr559', enabled_sensors=enabled_sensors, **kwargs)


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import dateutil.parser
import importlib

from abc import ABCMeta, abstractmethod

from platypush.plugins import Plugin, action


class CalendarInterface:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_upcoming_events(self, max_results=10):
        raise NotImplementedError()


class CalendarPlugin(Plugin, CalendarInterface):
    """
    The CalendarPlugin allows you to keep track of multiple calendars (Google or
    iCal URLs) and get joined events from all of them.

    Requires:
        * **dateutil** (``pip install python-dateutil``)
    """

    def __init__(self, calendars=None, *args, **kwargs):
        """
        :param calendars: List of calendars to be queried. Supported types so far: Google Calendar and iCal URLs.
        :type calendars: list

        Example format::

            calendars = [
                {
                    "type": "platypush.plugins.google.calendar.GoogleCalendarPlugin"
                },

                {
                    "type": "platypush.plugins.calendar.ical.IcalCalendarPlugin",
                    "url": "https://www.facebook.com/ical/u.php?uid=USER_ID&key=FB_KEY"
                },

                ...
            ]
        """

        super().__init__(*args, **kwargs)

        if calendars is None:
            calendars = []
        self.calendars = []

        for calendar in calendars:
            if 'type' not in calendar:
                self.logger.warning("Invalid calendar with no type specified: {}".format(calendar))
                continue

            cal_type = calendar.pop('type')
            module_name = '.'.join(cal_type.split('.')[:-1])
            class_name = cal_type.split('.')[-1]
            module = importlib.import_module(module_name)
            self.calendars.append(getattr(module, class_name)(**calendar))


    @action
    def get_upcoming_events(self, max_results=10):
        """
        Get a list of upcoming events merging all the available calendars.

        :param max_results: Maximum number of results to be returned (default: 10)
        :type max_results: int

        :returns: platypush.message.Response -- Response object with the list of events in the Google calendar API format.

        Example::

            output = [
                {
                    "id": "123456abcdef",
                    "kind": "calendar#event",
                    "status": "confirmed",
                    "htmlLink": "...",
                    "created": "2018-06-01T01:23:45.000Z",
                    "updated": "2018-06-01T01:23:45.000Z",
                    "creator": {
                        "email": "...",
                        "displayName": "...",
                        "self": true
                    },
                    "organizer" {
                        "email": "...",
                        "displayName": "...",
                        "self": true
                    },
                    "start": {
                        "dateTime": "2018-06-02T10:00:00.000Z",
                    },
                    "end": {
                        "dateTime": "2018-06-02T12:00:00.000Z",
                    },
                },
                ...
            ]
        """

        events = []

        for calendar in self.calendars:
            cal_events = calendar.get_upcoming_events().output or []
            events.extend(cal_events)

        events = sorted(events, key=lambda event:
                        dateutil.parser.parse(
                            event['start']['dateTime']
                            if 'dateTime' in event['start']
                            else event['start']['date'] + 'T00:00:00+00:00'
                        ))[:max_results]

        return events


# vim:sw=4:ts=4:et:




from typing import Optional

from platypush.message.event import Event


class BluetoothEvent(Event):
    pass


class BluetoothDeviceFoundEvent(Event):
    """
    Event triggered when a bluetooth device is found during a scan.
    """
    def __init__(self, address: str, name: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, address=address, name=name, **kwargs)


class BluetoothDeviceLostEvent(Event):
    """
    Event triggered when a bluetooth device previously scanned is lost.
    """
    def __init__(self, address: str, name: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, address=address, name=name, **kwargs)


class BluetoothDeviceConnectedEvent(Event):
    """
    Event triggered on bluetooth device connection
    """
    def __init__(self, address: str = None, port: str = None, *args, **kwargs):
        super().__init__(*args, address=address, port=port, **kwargs)


class BluetoothDeviceDisconnectedEvent(Event):
    """
    Event triggered on bluetooth device disconnection
    """
    def __init__(self, address: str = None, port: str = None, *args, **kwargs):
        super().__init__(*args, address=address, port=port, **kwargs)


class BluetoothConnectionRejectedEvent(Event):
    """
    Event triggered on bluetooth device connection rejected
    """
    def __init__(self, address: str = None, port: str = None, *args, **kwargs):
        super().__init__(*args, address=address, port=port, **kwargs)


class BluetoothFilePutRequestEvent(Event):
    """
    Event triggered on bluetooth device file transfer put request
    """
    def __init__(self, address: str = None, port: str = None, *args, **kwargs):
        super().__init__(*args, address=address, port=port, **kwargs)


class BluetoothFileGetRequestEvent(Event):
    """
    Event triggered on bluetooth device file transfer get request
    """
    def __init__(self, address: str = None, port: str = None, *args, **kwargs):
        super().__init__(*args, address=address, port=port, **kwargs)


class BluetoothFileReceivedEvent(Event):
    """
    Event triggered on bluetooth device file transfer put request
    """
    def __init__(self, path: str = None, *args, **kwargs):
        super().__init__(*args, path=path, **kwargs)


# vim:sw=4:ts=4:et:



import io
import os
import re
import shutil
import threading
import time

from datetime import datetime
from typing import Optional

from platypush.config import Config
from platypush.message import Mapping
from platypush.message.response import Response
from platypush.message.event.camera import CameraRecordingStartedEvent, \
    CameraRecordingStoppedEvent, CameraVideoRenderedEvent, \
    CameraPictureTakenEvent, CameraFrameCapturedEvent

from platypush.plugins import Plugin, action


class StreamingOutput:
    def __init__(self, raw=False):
        self.frame = None
        self.raw_frame = None
        self.raw = raw
        self.buffer = io.BytesIO()
        self.ready = threading.Condition()

    def is_new_frame(self, buf):
        if self.raw:
            return True

        # JPEG header begin
        return buf.startswith(b'\xff\xd8')

    def write(self, buf):
        if not self.is_new_frame(buf):
            return

        if self.raw:
            with self.ready:
                self.raw_frame = buf
                self.ready.notify_all()
                return

        # New frame, copy the existing buffer's content and notify all clients that it's available
        self.buffer.truncate()
        with self.ready:
            self.frame = self.buffer.getvalue()
            self.ready.notify_all()

        self.buffer.seek(0)
        return self.buffer.write(buf)

    def close(self):
        self.buffer.close()


class CameraPlugin(Plugin):
    """
    Plugin to control generic cameras over OpenCV.

    Triggers:

        * :class:`platypush.message.event.camera.CameraRecordingStartedEvent`
            when a new video recording/photo burst starts
        * :class:`platypush.message.event.camera.CameraRecordingStoppedEvent`
            when a video recording/photo burst ends
        * :class:`platypush.message.event.camera.CameraVideoRenderedEvent`
            when a sequence of captured is successfully rendered into a video
        * :class:`platypush.message.event.camera.CameraPictureTakenEvent`
            when a snapshot is captured and stored to an image file

    Requires:

        * **opencv** (``pip install opencv-python``)

    """

    _default_warmup_frames = 5
    _default_sleep_between_frames = 0
    _default_color_transform = 'COLOR_BGR2BGRA'
    _default_frames_dir = None

    _max_stored_frames = 100
    _frame_filename_regex = re.compile('(\d+)-(\d+)-(\d+)_(\d+)-(\d+)-(\d+)-(\d+).jpe?g$')

    def __init__(self, device_id=0, frames_dir=None,
                 warmup_frames=_default_warmup_frames, video_type=0,
                 sleep_between_frames=_default_sleep_between_frames,
                 max_stored_frames=_max_stored_frames,
                 color_transform=_default_color_transform,
                 scale_x=None, scale_y=None, rotate=None, flip=None, stream_raw_frames=False, **kwargs):
        """
        :param device_id: Index of the default video device to be used for
            capturing (default: 0)
        :type device_id: int

        :param frames_dir: Directory where the camera frames will be stored
            (default: ``~/.local/share/platypush/camera/frames``)
        :type frames_dir: str

        :param warmup_frames: Cameras usually take a while to adapt their
            luminosity and focus to the environment when taking a picture.
            This parameter allows you to specify the number of "warmup" frames
            to capture upon picture command before actually capturing a frame
            (default: 5 but you may want to calibrate this parameter for your
            camera)
        :type warmup_frames: int

        :param video_type: Default video type to use when exporting captured
            frames to camera (default: 0, infers the type from the video file
            extension). See
            `here <https://docs.opencv.org/4.0.1/dd/d9e/classcv_1_1VideoWriter.html#afec93f94dc6c0b3e28f4dd153bc5a7f0>`_
            for a reference on the supported types (e.g. 'MJPEG', 'XVID', 'H264' etc')
        :type video_type: str or int

        :param sleep_between_frames: If set, the process will sleep for the
            specified amount of seconds between two frames when recording
            (default: 0)
        :type sleep_between_frames: float

        :param max_stored_frames: Maximum number of frames to store in
            ``frames_dir`` when recording with no persistence (e.g. streaming
            over HTTP) (default: 100)
        :type max_stored_frames: int

        :param color_transform: Color transformation to apply to the captured
            frames. See https://docs.opencv.org/3.2.0/d7/d1b/group__imgproc__misc.html
            for a full list of supported color transformations.
            (default: "``COLOR_BGR2BGRA``")
        :type color_transform: str

        :param scale_x: If set, the images will be scaled along the x axis by the
            specified factor
        :type scale_x: float

        :param scale_y: If set, the images will be scaled along the y axis by the
            specified factor
        :type scale_y: float

        :param rotate: If set, the images will be rotated by the specified
            number of degrees
        :type rotate: float

        :param flip: If set, the images will be flipped around the specified
            axis. Possible values::

                - ``0`` - flip along the x axis
                - ``1`` - flip along the y axis
                - ``-1`` - flip along both the axis

        :type flip: int
        """

        super().__init__(**kwargs)

        self._default_frames_dir = os.path.join(Config.get('workdir'), 'camera', 'frames')
        self.default_device_id = device_id
        self.frames_dir = os.path.abspath(os.path.expanduser(frames_dir or self._default_frames_dir))
        self.warmup_frames = warmup_frames
        self.video_type = video_type
        self.stream_raw_frames = stream_raw_frames

        if isinstance(video_type, str):
            import cv2
            self.video_type = cv2.VideoWriter_fourcc(*video_type.upper())

        self.sleep_between_frames = sleep_between_frames
        self.max_stored_frames = max_stored_frames
        self.color_transform = color_transform
        self.scale_x = scale_x
        self.scale_y = scale_y
        self.rotate = rotate
        self.flip = flip

        self._is_recording = {}  # device_id => Event map
        self._devices = {}  # device_id => VideoCapture map
        self._recording_threads = {}  # device_id => Thread map
        self._recording_info = {}  # device_id => recording info map
        self._output = None

    def _init_device(self, device_id, frames_dir=None, **info):
        import cv2
        self._release_device(device_id)

        if device_id not in self._devices:
            self._devices[device_id] = cv2.VideoCapture(device_id)

        if device_id not in self._is_recording:
            self._is_recording[device_id] = threading.Event()

        self._recording_info[device_id] = info

        if frames_dir:
            os.makedirs(frames_dir, exist_ok=True)
            self._recording_info[device_id]['frames_dir'] = frames_dir

        return self._devices[device_id]

    def _release_device(self, device_id, wait_thread_termination=True):
        if device_id in self._is_recording:
            self._is_recording[device_id].clear()

        if device_id in self._recording_threads:
            if wait_thread_termination:
                self.logger.info('A recording thread is running, waiting for termination')
                if self._recording_threads[device_id].is_alive():
                    self._recording_threads[device_id].join()
                del self._recording_threads[device_id]

        if device_id in self._devices:
            self._devices[device_id].release()
            del self._devices[device_id]
            self.fire_event(CameraRecordingStoppedEvent(device_id=device_id))
            self.logger.info("Device {} released".format(device_id))

        if device_id in self._recording_info:
            del self._recording_info[device_id]

    @staticmethod
    def _store_frame_to_file(frame, frames_dir, image_file):
        import cv2

        if image_file:
            filepath = image_file
        else:
            filepath = os.path.join(
                frames_dir, datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f.jpg'))

        cv2.imwrite(filepath, frame)
        return filepath

    def _get_stored_frames_files(self, frames_dir):
        ret = sorted([
            os.path.join(frames_dir, f) for f in os.listdir(frames_dir)
            if os.path.isfile(os.path.join(frames_dir, f)) and
            re.search(self._frame_filename_regex, f)
        ])
        return ret

    def _get_avg_fps(self, frames_dir):
        files = self._get_stored_frames_files(frames_dir)
        frame_time_diff = 0.0
        n_frames = 0

        for i in range(1, len(files)):
            m1 = re.search(self._frame_filename_regex, files[i - 1])
            m2 = re.search(self._frame_filename_regex, files[i])

            if not m1 or not m2:
                continue

            t1 = datetime.timestamp(datetime(*map(int, m1.groups())))
            t2 = datetime.timestamp(datetime(*map(int, m2.groups())))
            frame_time_diff += (t2 - t1)
            n_frames += 1

        return n_frames / frame_time_diff if n_frames and frame_time_diff else 0

    def _remove_expired_frames(self, frames_dir, max_stored_frames):
        files = self._get_stored_frames_files(frames_dir)
        for f in files[:len(files) - max_stored_frames]:
            os.unlink(f)

    def _make_video_file(self, frames_dir, video_file, video_type):
        import cv2

        files = self._get_stored_frames_files(frames_dir)
        if not files:
            self.logger.warning('No frames found in {}'.format(frames_dir))
            return

        frame = cv2.imread(files[0])
        height, width, layers = frame.shape
        fps = self._get_avg_fps(frames_dir)
        video = cv2.VideoWriter(video_file, video_type, fps, (width, height))

        for f in files:
            video.write(cv2.imread(f))
        video.release()

        self.fire_event(CameraVideoRenderedEvent(filename=video_file))
        shutil.rmtree(frames_dir, ignore_errors=True)

    def _recording_thread(self):
        def thread(duration, video_file, image_file, device_id,
                   frames_dir, n_frames, sleep_between_frames,
                   max_stored_frames, color_transform, video_type,
                   scale_x, scale_y, rotate, flip):
            import cv2
            device = self._devices[device_id]
            color_transform = getattr(cv2, color_transform or self.color_transform)
            rotation_matrix = None
            self._is_recording[device_id].wait()
            self.logger.info('Starting recording from video device {}'.format(device_id))
            recording_started_time = time.time()
            captured_frames = 0

            evt_args = {
                'device_id': device_id,
            }

            if video_file or image_file:
                evt_args['filename'] = video_file or image_file
            if frames_dir:
                evt_args['frames_dir'] = frames_dir

            self.fire_event(CameraRecordingStartedEvent(**evt_args))

            while device_id in self._is_recording and self._is_recording[device_id].is_set():
                if duration and time.time() - recording_started_time >= duration \
                        or n_frames and captured_frames >= n_frames:
                    break

                ret, frame = device.read()
                if not ret:
                    self.logger.warning('Error while retrieving video frame')
                    continue

                frame = cv2.cvtColor(frame, color_transform)

                if rotate:
                    rows, cols = frame.shape
                    if not rotation_matrix:
                        rotation_matrix = cv2.getRotationMatrix2D((cols / 2, rows / 2), rotate, 1)

                    frame = cv2.warpAffine(frame, rotation_matrix, (cols, rows))

                if flip is not None:
                    frame = cv2.flip(frame, flip)

                if scale_x or scale_y:
                    scale_x = scale_x or 1
                    scale_y = scale_y or 1
                    frame = cv2.resize(frame, None, fx=scale_x, fy=scale_y,
                                       interpolation=cv2.INTER_CUBIC)

                if self._output:
                    if not self.stream_raw_frames:
                        result, frame = cv2.imencode('.jpg', frame)
                        if not result:
                            self.logger.warning('Unable to convert frame to JPEG')
                            continue

                        self._output.write(frame.tobytes())
                    else:
                        self._output.write(frame)
                elif frames_dir:
                    self._store_frame_to_file(frame=frame, frames_dir=frames_dir, image_file=image_file)

                captured_frames += 1
                self.fire_event(CameraFrameCapturedEvent(filename=image_file))

                if max_stored_frames and not video_file:
                    self._remove_expired_frames(
                        frames_dir=frames_dir,
                        max_stored_frames=max_stored_frames)

                if sleep_between_frames:
                    time.sleep(sleep_between_frames)

            self._release_device(device_id, wait_thread_termination=False)

            if image_file:
                self.fire_event(CameraPictureTakenEvent(filename=image_file))

            self.logger.info('Recording terminated')

            if video_file:
                self.logger.info('Writing frames to video file {}'.
                                 format(video_file))
                self._make_video_file(frames_dir=frames_dir,
                                      video_file=video_file,
                                      video_type=video_type)
                self.logger.info('Video file {}: rendering completed'.
                                 format(video_file))

        return thread

    @action
    def start_recording(self, duration: Optional[float] = None, video_file: Optional[str] = None,
                        video_type: Optional[str] = None, device_id: Optional[int] = None,
                        frames_dir: Optional[str] = None, sleep_between_frames: Optional[float] = None,
                        max_stored_frames: Optional[int] = None, color_transform: Optional[str] = None,
                        scale_x: Optional[float] = None, scale_y: Optional[float] = None,
                        rotate: Optional[float] = None, flip: Optional[int] = None):
        """
        Start recording

        :param duration: Record duration in seconds (default: None, record until
            ``stop_recording``)
        :param video_file: If set, the stream will be recorded to the specified
            video file (default: None)
        :param video_type: Overrides the default configured ``video_type``

        :param device_id: Override default device_id
        :param frames_dir: Override default frames_dir
        :param sleep_between_frames: Override default sleep_between_frames
        :param max_stored_frames: Override default max_stored_frames
        :param color_transform: Override default color_transform
        :param scale_x: Override default scale_x
        :param scale_y: Override default scale_y
        :param rotate: Override default rotate
        :param flip: Override default flip
        """

        device_id = device_id if device_id is not None else self.default_device_id
        if device_id in self._is_recording and \
                self._is_recording[device_id].is_set():
            self.logger.info('A recording on device {} is already in progress'.
                             format(device_id))
            return self.status(device_id=device_id)

        recording_started = threading.Event()

        # noinspection PyUnusedLocal
        def on_recording_started(event):
            recording_started.set()

        attrs = self._get_attributes(frames_dir=frames_dir, sleep_between_frames=sleep_between_frames,
                                     max_stored_frames=max_stored_frames, color_transform=color_transform,
                                     scale_x=scale_x, scale_y=scale_y, rotate=rotate, flip=flip, video_type=video_type)

        # noinspection PyUnresolvedReferences
        if attrs.frames_dir:
            # noinspection PyUnresolvedReferences
            attrs.frames_dir = os.path.join(attrs.frames_dir, str(device_id))
            if video_file:
                video_file = os.path.abspath(os.path.expanduser(video_file))
                attrs.frames_dir = os.path.join(attrs.frames_dir, 'recording_{}'.format(
                    datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')))

        # noinspection PyUnresolvedReferences
        self._init_device(device_id,
                          video_file=video_file,
                          video_type=attrs.video_type,
                          frames_dir=attrs.frames_dir,
                          sleep_between_frames=attrs.sleep_between_frames,
                          max_stored_frames=attrs.max_stored_frames,
                          color_transform=attrs.color_transform,
                          scale_x=attrs.scale_x,
                          scale_y=attrs.scale_y,
                          rotate=attrs.rotate,
                          flip=attrs.flip)

        self.register_handler(CameraRecordingStartedEvent, on_recording_started)

        # noinspection PyUnresolvedReferences
        self._recording_threads[device_id] = threading.Thread(
            target=self._recording_thread(), kwargs={
                'duration': duration,
                'video_file': video_file,
                'video_type': attrs.video_type,
                'image_file': None,
                'device_id': device_id,
                'frames_dir': attrs.frames_dir,
                'n_frames': None,
                'sleep_between_frames': attrs.sleep_between_frames,
                'max_stored_frames': attrs.max_stored_frames,
                'color_transform': attrs.color_transform,
                'scale_x': attrs.scale_x,
                'scale_y': attrs.scale_y,
                'rotate': attrs.rotate,
                'flip': attrs.flip,
            })

        self._recording_threads[device_id].start()
        self._is_recording[device_id].set()

        recording_started.wait()
        self.unregister_handler(CameraRecordingStartedEvent, on_recording_started)
        return self.status(device_id=device_id)

    @action
    def stop_recording(self, device_id=None):
        """
        Stop recording
        """

        device_id = device_id if device_id is not None else self.default_device_id
        frames_dir = self._recording_info.get(device_id, {}).get('frames_dir')
        self._release_device(device_id)
        shutil.rmtree(frames_dir, ignore_errors=True)

    def _get_attributes(self, frames_dir=None, warmup_frames=None,
                        color_transform=None, scale_x=None, scale_y=None,
                        rotate=None, flip=None, sleep_between_frames=None,
                        max_stored_frames=None, video_type=None) -> Mapping:
        import cv2

        warmup_frames = warmup_frames if warmup_frames is not None else self.warmup_frames
        frames_dir = os.path.abspath(os.path.expanduser(frames_dir)) if frames_dir is not None else self.frames_dir
        sleep_between_frames = sleep_between_frames if sleep_between_frames is not None else self.sleep_between_frames
        max_stored_frames = max_stored_frames if max_stored_frames is not None else self.max_stored_frames
        color_transform = color_transform if color_transform is not None else self.color_transform
        scale_x = scale_x if scale_x is not None else self.scale_x
        scale_y = scale_y if scale_y is not None else self.scale_y
        rotate = rotate if rotate is not None else self.rotate
        flip = flip if flip is not None else self.flip
        if video_type is not None:
            video_type = cv2.VideoWriter_fourcc(*video_type.upper()) if isinstance(video_type, str) else video_type
        else:
            video_type = self.video_type

        return Mapping(warmup_frames=warmup_frames, frames_dir=frames_dir, sleep_between_frames=sleep_between_frames,
                       max_stored_frames=max_stored_frames, color_transform=color_transform, scale_x=scale_x,
                       scale_y=scale_y, rotate=rotate, flip=flip, video_type=video_type)

    @action
    def take_picture(self, image_file: str, device_id: Optional[int] = None, warmup_frames: Optional[int] = None,
                     color_transform: Optional[str] = None, scale_x: Optional[float] = None,
                     scale_y: Optional[float] = None, rotate: Optional[float] = None, flip: Optional[int] = None):
        """
        Take a picture.

        :param image_file: Path where the output image will be stored.
        :param device_id: Override default device_id
        :param warmup_frames: Override default warmup_frames
        :param color_transform: Override default color_transform
        :param scale_x: Override default scale_x
        :param scale_y: Override default scale_y
        :param rotate: Override default rotate
        :param flip: Override default flip
        """

        device_id = device_id if device_id is not None else self.default_device_id
        image_file = os.path.abspath(os.path.expanduser(image_file))
        picture_taken = threading.Event()

        # noinspection PyUnusedLocal
        def on_picture_taken(event):
            picture_taken.set()

        if device_id in self._is_recording and \
                self._is_recording[device_id].is_set():
            self.logger.info('A recording on device {} is already in progress'.
                             format(device_id))

            status = self.status(device_id=device_id).output.get(device_id)
            if 'image_file' in status:
                shutil.copyfile(status['image_file'], image_file)
                return {'path': image_file}

            raise RuntimeError('Recording already in progress and no images ' +
                               'have been captured yet')

        attrs = self._get_attributes(warmup_frames=warmup_frames, color_transform=color_transform, scale_x=scale_x,
                                     scale_y=scale_y, rotate=rotate, flip=flip)

        # noinspection PyUnresolvedReferences
        self._init_device(device_id, image_file=image_file, warmup_frames=attrs.warmup_frames,
                          color_transform=attrs.color_transform, scale_x=attrs.scale_x, scale_y=attrs.scale_y,
                          rotate=attrs.rotate, flip=attrs.flip)

        self.register_handler(CameraPictureTakenEvent, on_picture_taken)
        self._recording_threads[device_id] = threading.Thread(
            target=self._recording_thread(), kwargs={
                'duration': None, 'video_file': None,
                'image_file': image_file, 'video_type': None,
                'device_id': device_id, 'frames_dir': None,
                'n_frames': warmup_frames,
                'sleep_between_frames': None,
                'max_stored_frames': None,
                'color_transform': color_transform,
                'scale_x': scale_x, 'scale_y': scale_y,
                'rotate': rotate, 'flip': flip
            })

        self._recording_threads[device_id].start()
        self._is_recording[device_id].set()

        picture_taken.wait()
        self.unregister_handler(CameraPictureTakenEvent, on_picture_taken)
        return {'path': image_file}

    @action
    def status(self, device_id=None):
        """
        Returns the status of the specified device_id or all the device in a
        ``{ device_id => device_info }`` map format. Device info includes
        ``video_file``, ``image_file``, ``frames_dir`` and additional video info
        """

        resp = Response(output={
            id: {
                'image_file': self._get_stored_frames_files(info['frames_dir'])[-2]
                if 'frames_dir' in info
                   and len(self._get_stored_frames_files(info['frames_dir'])) > 1
                   and 'image_file' not in info else info.get('image_file'), **info
            }
            for id, info in self._recording_info.items()
            if device_id is None or id == device_id
        }, disable_logging=True)
        return resp

    @action
    def get_default_device_id(self):
        return self.default_device_id

    def get_stream(self):
        return self._output

    def __enter__(self):
        device_id = self.default_device_id
        self._output = StreamingOutput(raw=self.stream_raw_frames)
        self._init_device(device_id=device_id)
        self.start_recording(device_id=device_id)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_recording(self.default_device_id)
        if self._output:
            self._output.close()
        self._output = None


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class PingEvent(Event):
    """ Ping event, used for testing purposes """

    def __init__(self, message=None, *args, **kwargs):
        """
        :param message: Ping message
        :type message: object
        """

        super().__init__(message=message, *args, **kwargs)


class HostDownEvent(Event):
    """
    Event triggered when a remote host stops responding ping requests.
    """
    def __init__(self, host: str, *args, **kwargs):
        super().__init__(host=host, *args, **kwargs)


class HostUpEvent(Event):
    """
    Event triggered when a remote host starts responding ping requests.
    """
    def __init__(self, host: str, *args, **kwargs):
        super().__init__(host=host, *args, **kwargs)


# vim:sw=4:ts=4:et:



import threading
import time

from platypush.backend import Backend
from platypush.message.event.gps import GPSVersionEvent, GPSDeviceEvent, GPSUpdateEvent


class GpsBackend(Backend):
    """
    This backend can interact with a GPS device and listen for events.

    Triggers:

        * :class:`platypush.message.event.gps.GPSVersionEvent` when a GPS device advertises its version data
        * :class:`platypush.message.event.gps.GPSDeviceEvent` when a GPS device is connected or updated
        * :class:`platypush.message.event.gps.GPSUpdateEvent` when a GPS device has new data

    Requires:

        * **gps** (``pip install gps``)
        * **gpsd** daemon running (``apt-get install gpsd`` or ``pacman -S gpsd`` depending on your distro)

    Once installed gpsd you need to run it and associate it to your device. Example if your GPS device communicates
    over USB and is available on /dev/ttyUSB0::

        [sudo] gpsd /dev/ttyUSB0 -F /var/run/gpsd.sock

    The best option is probably to run gpsd at startup as a systemd service.
    """

    _fail_sleep_time = 5.0
    _lat_lng_tolerance = 1e-5
    _alt_tolerance = 0.5

    def __init__(self, gpsd_server='localhost', gpsd_port=2947, **kwargs):
        """
        :param gpsd_server: gpsd daemon server name/address (default: localhost)
        :type gpsd_server: str
        :param gpsd_port: Port of the gpsd daemon (default: 2947)
        :type gpsd_port: int or str
        """
        super().__init__(**kwargs)

        self.gpsd_server = gpsd_server
        self.gpsd_port = gpsd_port
        self._session = None
        self._session_lock = threading.RLock()
        self._devices = {}

    def _get_session(self):
        import gps

        with self._session_lock:
            if not self._session:
                self._session = gps.gps(host=self.gpsd_server, port=self.gpsd_port, reconnect=True)
                self._session.stream(gps.WATCH_ENABLE | gps.WATCH_NEWSTYLE)

        return self._session

    def _gps_report_to_event(self, report):
        if report.get('class').lower() == 'version':
            return GPSVersionEvent(release=report.get('release'),
                                   rev=report.get('rev'),
                                   proto_major=report.get('proto_major'),
                                   proto_minor=report.get('proto_minor'))
        if report.get('class').lower() == 'devices':
            for device in report.get('devices', []):
                if device.get('path') not in self._devices or device != self._devices.get('path'):
                    self._devices[device.get('path')] = device
                    return GPSDeviceEvent(path=device.get('path'), activated=device.get('activated'),
                                          native=device.get('native'), bps=device.get('bps'),
                                          parity=device.get('parity'), stopbits=device.get('stopbits'),
                                          cycle=device.get('cycle'), driver=device.get('driver'))
        if report.get('class').lower() == 'device':
            self._devices[report.get('path')] = report
            return GPSDeviceEvent(path=report.get('path'), activated=report.get('activated'),
                                  native=report.get('native'), bps=report.get('bps'),
                                  parity=report.get('parity'), stopbits=report.get('stopbits'),
                                  cycle=report.get('cycle'), driver=report.get('driver'))
        if report.get('class').lower() == 'tpv':
            return GPSUpdateEvent(device=report.get('device'), latitude=report.get('lat'), longitude=report.get('lon'),
                                  altitude=report.get('alt'), mode=report.get('mode'), epv=report.get('epv'),
                                  eph=report.get('eph'), sep=report.get('sep'))

    def run(self):
        super().run()
        self.logger.info('Initialized GPS backend on {}:{}'.format(self.gpsd_server, self.gpsd_port))
        last_event = None

        while not self.should_stop():
            try:
                session = self._get_session()
                report = session.next()
                event = self._gps_report_to_event(report)
                if event and (last_event is None or
                              abs((last_event.args.get('latitude') or 0) - (event.args.get('latitude') or 0)) >= self._lat_lng_tolerance or
                              abs((last_event.args.get('longitude') or 0) - (event.args.get('longitude') or 0)) >= self._lat_lng_tolerance or
                              abs((last_event.args.get('altitude') or 0) - (event.args.get('altitude') or 0)) >= self._alt_tolerance):
                    self.bus.post(event)
                    last_event = event
            except Exception as e:
                if isinstance(e, StopIteration):
                    self.logger.warning('GPS service connection lost, check that gpsd is running')
                else:
                    self.logger.exception(e)

                self._session = None
                time.sleep(self._fail_sleep_time)


# vim:sw=4:ts=4:et:



import asyncio
import importlib
import logging

from threading import RLock

from ..config import Config

logger = logging.getLogger(__name__)

# Map: backend_name -> backend_instance
backends = {}

# Map: plugin_name -> plugin_instance
plugins = {}

# Map: plugin_name -> init_lock to make sure that a plugin isn't initialized
# multiple times
plugins_init_locks = {}

# Reference to the main application bus
main_bus = None

def register_backends(bus=None, global_scope=False, **kwargs):
    """ Initialize the backend objects based on the configuration and returns
        a name -> backend_instance map.
    Params:
        bus -- If specific (it usually should), the messages processed by the
            backends will be posted on this bus.

        kwargs -- Any additional key-value parameters required to initialize the backends
        """

    global main_bus
    if bus:
        main_bus = bus

    if global_scope:
        global backends
    else:
        backends = {}

    for (name, cfg) in Config.get_backends().items():
        module = importlib.import_module('platypush.backend.' + name)

        # e.g. backend.pushbullet main class: PushbulletBackend
        cls_name = ''
        for token in module.__name__.title().split('.')[2:]:
            cls_name += token.title()
        cls_name += 'Backend'

        try:
            b = getattr(module, cls_name)(bus=bus, **cfg, **kwargs)
            backends[name] = b
        except AttributeError as e:
            logger.warning('No such class in {}: {}'.format(
                module.__name__, cls_name))
            raise RuntimeError(e)

    return backends

def get_backend(name):
    """ Returns the backend instance identified by name if it exists """

    global backends
    return backends.get(name)


def get_plugin(plugin_name, reload=False):
    """ Registers a plugin instance by name if not registered already, or
        returns the registered plugin instance"""
    global plugins
    global plugins_init_locks

    if plugin_name not in plugins_init_locks:
        plugins_init_locks[plugin_name] = RLock()

    if plugin_name in plugins and not reload:
        return plugins[plugin_name]

    try:
        plugin = importlib.import_module('platypush.plugins.' + plugin_name)
    except ImportError as e:
        logger.warning('No such plugin: {}'.format(plugin_name))
        raise RuntimeError(e)

    # e.g. plugins.music.mpd main class: MusicMpdPlugin
    cls_name = ''
    for token in plugin_name.split('.'):
        cls_name += token.title()
    cls_name += 'Plugin'

    plugin_conf = Config.get_plugins()[plugin_name] \
        if plugin_name in Config.get_plugins() else {}

    if 'disabled' in plugin_conf:
        if plugin_conf['disabled'] is True:
            return None
        del plugin_conf['disabled']

    if 'enabled' in plugin_conf:
        if plugin_conf['enabled'] is False:
            return None
        del plugin_conf['enabled']

    try:
        plugin_class = getattr(plugin, cls_name)
    except AttributeError as e:
        logger.warning('No such class in {}: {} [error: {}]'.format(plugin_name, cls_name, str(e)))
        raise RuntimeError(e)

    with plugins_init_locks[plugin_name]:
        if plugins.get(plugin_name) and not reload:
            return plugins[plugin_name]
        plugins[plugin_name] = plugin_class(**plugin_conf)

    return plugins[plugin_name]


def get_bus():
    global main_bus
    return main_bus


def get_or_create_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    return loop


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorMcp3008Backend(SensorBackend):
    """
    Backend to poll analog sensor values from an MCP3008 chipset
    (https://learn.adafruit.com/raspberry-pi-analog-to-digital-converters/mcp3008)

    Requires:

        * ``adafruit-mcp3008`` (``pip install adafruit-mcp3008``)
        * The :mod:`platypush.plugins.gpio.sensor.mcp3008` plugin configured
    """

    def __init__(self, **kwargs):
        super().__init__(plugin='gpio.sensor.mcp3008', **kwargs)


# vim:sw=4:ts=4:et:



#!/usr/bin/env python
# coding: latin-1
"""
This module is designed to communicate with the ZeroBorg

Use by creating an instance of the class, call the Init function, then command as desired, e.g.
import ZeroBorg
ZB = ZeroBorg.ZeroBorg()
ZB.Init()
# User code here, use ZB to control the board

Multiple boards can be used when configured with different I?C addresses by creating multiple instances, e.g.
import ZeroBorg
ZB1 = ZeroBorg.ZeroBorg()
ZB2 = ZeroBorg.ZeroBorg()
ZB1.i2cAddress = 0x44
ZB2.i2cAddress = 0x45
ZB1.Init()
ZB2.Init()
# User code here, use ZB1 and ZB2 to control each board separately

For explanations of the functions available call the Help function, e.g.
import ZeroBorg
ZB = ZeroBorg.ZeroBorg()
ZB.Help()
See the website at www.piborg.org/zeroborg for more details
"""

# Import the libraries we need
import io
import fcntl
import types
import time

# Constant values
I2C_SLAVE = 0x0703
PWM_MAX = 255
I2C_NORM_LEN = 4
I2C_LONG_LEN = 24

I2C_ID_ZEROBORG = 0x40

COMMAND_SET_LED = 1  # Set the LED status
COMMAND_GET_LED = 2  # Get the LED status
COMMAND_SET_A_FWD = 3  # Set motor 1 PWM rate in a forwards direction
COMMAND_SET_A_REV = 4  # Set motor 1 PWM rate in a reverse direction
COMMAND_GET_A = 5  # Get motor 1 direction and PWM rate
COMMAND_SET_B_FWD = 6  # Set motor 2 PWM rate in a forwards direction
COMMAND_SET_B_REV = 7  # Set motor 2 PWM rate in a reverse direction
COMMAND_GET_B = 8  # Get motor 2 direction and PWM rate
COMMAND_SET_C_FWD = 9  # Set motor 3 PWM rate in a forwards direction
COMMAND_SET_C_REV = 10  # Set motor 3 PWM rate in a reverse direction
COMMAND_GET_C = 11  # Get motor 3 direction and PWM rate
COMMAND_SET_D_FWD = 12  # Set motor 4 PWM rate in a forwards direction
COMMAND_SET_D_REV = 13  # Set motor 4 PWM rate in a reverse direction
COMMAND_GET_D = 14  # Get motor 4 direction and PWM rate
COMMAND_ALL_OFF = 15  # Switch everything off
COMMAND_SET_ALL_FWD = 16  # Set all motors PWM rate in a forwards direction
COMMAND_SET_ALL_REV = 17  # Set all motors PWM rate in a reverse direction
COMMAND_SET_FAILSAFE = 18  # Set the failsafe flag, turns the motors off if communication is interrupted
COMMAND_GET_FAILSAFE = 19  # Get the failsafe flag
COMMAND_RESET_EPO = 20  # Resets the EPO flag, use after EPO has been tripped and switch is now clear
COMMAND_GET_EPO = 21  # Get the EPO latched flag
COMMAND_SET_EPO_IGNORE = 22  # Set the EPO ignored flag, allows the system to run without an EPO
COMMAND_GET_EPO_IGNORE = 23  # Get the EPO ignored flag
COMMAND_GET_NEW_IR = 24  # Get the new IR message received flag
COMMAND_GET_LAST_IR = 25  # Get the last IR message received (long message, resets new IR flag)
COMMAND_SET_LED_IR = 26  # Set the LED for indicating IR messages
COMMAND_GET_LED_IR = 27  # Get if the LED is being used to indicate IR messages
COMMAND_GET_ANALOG_1 = 28  # Get the analog reading from port #1, pin 2
COMMAND_GET_ANALOG_2 = 29  # Get the analog reading from port #2, pin 4
COMMAND_GET_ID = 0x99  # Get the board identifier
COMMAND_SET_I2C_ADD = 0xAA  # Set a new I2C address

COMMAND_VALUE_FWD = 1  # I2C value representing forward
COMMAND_VALUE_REV = 2  # I2C value representing reverse

COMMAND_VALUE_ON = 1  # I2C value representing on
COMMAND_VALUE_OFF = 0  # I2C value representing off

COMMAND_ANALOG_MAX = 0x3FF  # Maximum value for analog readings

IR_MAX_BYTES = I2C_LONG_LEN - 2


# noinspection PyPep8Naming
def ScanForZeroBorg(busNumber=1):
    """
ScanForZeroBorg([busNumber])

Scans the I?C bus for a ZeroBorg boards and returns a list of all usable addresses
The busNumber if supplied is which I?C bus to scan, 0 for Rev 1 boards, 1 for Rev 2 boards, if not supplied
the default is 1
    """
    found = []
    print('Scanning I?C bus #%d' % busNumber)
    bus = ZeroBorg()
    for address in range(0x03, 0x78, 1):
        try:
            bus.InitBusOnly(busNumber, address)
            i2cRecv = bus.RawRead(COMMAND_GET_ID, I2C_NORM_LEN)
            if len(i2cRecv) == I2C_NORM_LEN:
                if i2cRecv[1] == I2C_ID_ZEROBORG:
                    print('Found ZeroBorg at %02X' % address)
                    found.append(address)
                else:
                    pass
            else:
                pass
        except KeyboardInterrupt:
            raise
        except Exception as e:
            print('Error on ZeroBorg scan: {}: {}'.format(type(e), str(e)))

    if len(found) == 0:
        print('No ZeroBorg boards found, is bus #%d correct (should be 0 for Rev 1, 1 for Rev 2)' % busNumber)
    elif len(found) == 1:
        print('1 ZeroBorg board found')
    else:
        print('%d ZeroBorg boards found' % (len(found)))
    return found


# noinspection PyPep8Naming
def SetNewAddress(newAddress, oldAddress=-1, busNumber=1):
    """
SetNewAddress(newAddress, [oldAddress], [busNumber])

Scans the I?C bus for the first ZeroBorg and sets it to a new I2C address
If oldAddress is supplied it will change the address of the board at that address rather than scanning the bus
The busNumber if supplied is which I?C bus to scan, 0 for Rev 1 boards, 1 for Rev 2 boards, if not supplied
the default is 1.
Warning, this new I?C address will still be used after resetting the power on the device
    """
    if newAddress < 0x03:
        print('Error, I?C addresses below 3 (0x03) are reserved, use an address between 3 (0x03) and 119 (0x77)')
        return
    elif newAddress > 0x77:
        print('Error, I?C addresses above 119 (0x77) are reserved, use an address between 3 (0x03) and 119 (0x77)')
        return
    if oldAddress < 0x0:
        found = ScanForZeroBorg(busNumber)
        if len(found) < 1:
            print('No ZeroBorg boards found, cannot set a new I?C address!')
            return
        else:
            oldAddress = found[0]
    print('Changing I2C address from %02X to %02X (bus #%d)' % (oldAddress, newAddress, busNumber))
    bus = ZeroBorg()
    bus.InitBusOnly(busNumber, oldAddress)
    # noinspection PyBroadException
    try:
        i2cRecv = bus.RawRead(COMMAND_GET_ID, I2C_NORM_LEN)
        if len(i2cRecv) == I2C_NORM_LEN:
            if i2cRecv[1] == I2C_ID_ZEROBORG:
                foundChip = True
                print('Found ZeroBorg at %02X' % oldAddress)
            else:
                foundChip = False
                print('Found a device at %02X, but it is not a ZeroBorg (ID %02X instead of %02X)' % (
                    oldAddress, i2cRecv[1], I2C_ID_ZEROBORG))
        else:
            foundChip = False
            print('Missing ZeroBorg at %02X' % oldAddress)
    except KeyboardInterrupt:
        raise
    except:
        foundChip = False
        print('Missing ZeroBorg at %02X' % oldAddress)
    if foundChip:
        bus.RawWrite(COMMAND_SET_I2C_ADD, [newAddress])
        time.sleep(0.1)
        print('Address changed to %02X, attempting to talk with the new address' % newAddress)
        # noinspection PyBroadException
        try:
            bus.InitBusOnly(busNumber, newAddress)
            i2cRecv = bus.RawRead(COMMAND_GET_ID, I2C_NORM_LEN)
            if len(i2cRecv) == I2C_NORM_LEN:
                if i2cRecv[1] == I2C_ID_ZEROBORG:
                    foundChip = True
                    print('Found ZeroBorg at %02X' % newAddress)
                else:
                    foundChip = False
                    print('Found a device at %02X, but it is not a ZeroBorg (ID %02X instead of %02X)' % (
                        newAddress, i2cRecv[1], I2C_ID_ZEROBORG))
            else:
                foundChip = False
                print('Missing ZeroBorg at %02X' % newAddress)
        except KeyboardInterrupt:
            raise
        except:
            foundChip = False
            print('Missing ZeroBorg at %02X' % newAddress)
    if foundChip:
        print('New I?C address of %02X set successfully' % newAddress)
    else:
        print('Failed to set new I?C address...')


# Class used to control ZeroBorg
# noinspection PyPep8Naming
class ZeroBorg:
    """
This module is designed to communicate with the ZeroBorg

busNumber               I?C bus on which the ZeroBorg is attached (Rev 1 is bus 0, Rev 2 is bus 1)
bus                     the smbus object used to talk to the I?C bus
i2cAddress              The I?C address of the ZeroBorg chip to control
foundChip               True if the ZeroBorg chip can be seen, False otherwise
printFunction           Function reference to call when printing text, if None "print" is used
    """

    # Shared values used by this class
    busNumber = 1  # Check here for Rev 1 vs Rev 2 and select the correct bus
    i2cAddress = I2C_ID_ZEROBORG  # I?C address, override for a different address
    foundChip = False
    printFunction = None
    i2cWrite = None
    i2cRead = None

    def RawWrite(self, command, data):
        """
RawWrite(command, data)

Sends a raw command on the I2C bus to the ZeroBorg
Command codes can be found at the top of ZeroBorg.py, data is a list of 0 or more byte values

Under most circumstances you should use the appropriate function instead of RawWrite
        """
        rawOutput = bytes([command])
        if data:
            rawOutput += bytes(data)
        self.i2cWrite.write(rawOutput)

    def RawRead(self, command, length, retryCount=3):
        """
RawRead(command, length, [retryCount])

Reads data back from the ZeroBorg after sending a GET command
Command codes can be found at the top of ZeroBorg.py, length is the number of bytes to read back

The function checks that the first byte read back matches the requested command
If it does not it will retry the request until retryCount is exhausted (default is 3 times)

Under most circumstances you should use the appropriate function instead of RawRead
        """
        reply = []

        while retryCount > 0:
            self.RawWrite(command, [])
            rawReply = self.i2cRead.read(length)
            reply = []
            for singleByte in rawReply:
                reply.append(singleByte)
            if command == reply[0]:
                break
            else:
                retryCount -= 1
        if retryCount > 0:
            return reply
        else:
            raise IOError('I2C read for command %d failed' % command)

    def InitBusOnly(self, busNumber, address):
        """
InitBusOnly(busNumber, address)

Prepare the I2C driver for talking to a ZeroBorg on the specified bus and I2C address
This call does not check the board is present or working, under most circumstances use Init() instead
        """
        self.busNumber = busNumber
        self.i2cAddress = address
        self.i2cRead = io.open("/dev/i2c-" + str(self.busNumber), "rb", buffering=0)
        fcntl.ioctl(self.i2cRead, I2C_SLAVE, self.i2cAddress)
        self.i2cWrite = io.open("/dev/i2c-" + str(self.busNumber), "wb", buffering=0)
        fcntl.ioctl(self.i2cWrite, I2C_SLAVE, self.i2cAddress)

    def Print(self, message):
        """
Print(message)

Wrapper used by the ZeroBorg instance to print messages, will call printFunction if set, print otherwise
        """
        if self.printFunction is None:
            print(message)
        else:
            self.printFunction(message)

    def NoPrint(self, message):
        """
NoPrint(message)

Does nothing, intended for disabling diagnostic printout by using:
ZB = ZeroBorg.ZeroBorg()
ZB.printFunction = ZB.NoPrint
        """
        pass

    def Init(self, tryOtherBus=False):
        """
Init([tryOtherBus])

Prepare the I2C driver for talking to the ZeroBorg

If tryOtherBus is True, this function will attempt to use the other bus if the ThunderBorg devices can not be found on
the current busNumber.

    This is only really useful for early Raspberry Pi models!
        """
        self.Print('Loading ZeroBorg on bus %d, address %02X' % (self.busNumber, self.i2cAddress))

        # Open the bus
        self.i2cRead = io.open("/dev/i2c-" + str(self.busNumber), "rb", buffering=0)
        fcntl.ioctl(self.i2cRead, I2C_SLAVE, self.i2cAddress)
        self.i2cWrite = io.open("/dev/i2c-" + str(self.busNumber), "wb", buffering=0)
        fcntl.ioctl(self.i2cWrite, I2C_SLAVE, self.i2cAddress)

        # Check for ZeroBorg
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_ID, I2C_NORM_LEN)
            if len(i2cRecv) == I2C_NORM_LEN:
                if i2cRecv[1] == I2C_ID_ZEROBORG:
                    self.foundChip = True
                    self.Print('Found ZeroBorg at %02X' % self.i2cAddress)
                else:
                    self.foundChip = False
                    self.Print('Found a device at %02X, but it is not a ZeroBorg (ID %02X instead of %02X)' % (
                        self.i2cAddress, i2cRecv[1], I2C_ID_ZEROBORG))
            else:
                self.foundChip = False
                self.Print('Missing ZeroBorg at %02X' % self.i2cAddress)
        except KeyboardInterrupt:
            raise
        except:
            self.foundChip = False
            self.Print('Missing ZeroBorg at %02X' % self.i2cAddress)

        # See if we are missing chips
        if not self.foundChip:
            self.Print('ZeroBorg was not found')
            if tryOtherBus:
                if self.busNumber == 1:
                    self.busNumber = 0
                else:
                    self.busNumber = 1
                self.Print('Trying bus %d instead' % self.busNumber)
                self.Init(False)
            else:
                self.Print(
                    'Are you sure your ZeroBorg is properly attached, the correct address is used, and the I2C ' +
                    'drivers are running?')
                self.bus = None
        else:
            self.Print('ZeroBorg loaded on bus %d' % (self.busNumber))

    def SetMotor1(self, power):
        """
SetMotor1(power)

Sets the drive level for motor 1, from +1 to -1.
e.g.
SetMotor1(0)     -> motor 1 is stopped
SetMotor1(0.75)  -> motor 1 moving forward at 75% power
SetMotor1(-0.5)  -> motor 1 moving reverse at 50% power
SetMotor1(1)     -> motor 1 moving forward at 100% power
        """
        if power < 0:
            # Reverse
            command = COMMAND_SET_A_REV
            pwm = -int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX
        else:
            # Forward / stopped
            command = COMMAND_SET_A_FWD
            pwm = int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX

        # noinspection PyBroadException
        try:
            self.RawWrite(command, [pwm])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motor 1 drive level!')

    def GetMotor1(self):
        """
power = GetMotor1()

Gets the drive level for motor 1, from +1 to -1.
e.g.
0     -> motor 1 is stopped
0.75  -> motor 1 moving forward at 75% power
-0.5  -> motor 1 moving reverse at 50% power
1     -> motor 1 moving forward at 100% power
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_A, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading motor 1 drive level!')
            return

        power = float(i2cRecv[2]) / float(PWM_MAX)

        if i2cRecv[1] == COMMAND_VALUE_FWD:
            return power
        elif i2cRecv[1] == COMMAND_VALUE_REV:
            return -power
        else:
            return

    def SetMotor2(self, power):
        """
SetMotor1(power)

Sets the drive level for motor 2, from +1 to -1.
e.g.
SetMotor2(0)     -> motor 2 is stopped
SetMotor2(0.75)  -> motor 2 moving forward at 75% power
SetMotor2(-0.5)  -> motor 2 moving reverse at 50% power
SetMotor2(1)     -> motor 2 moving forward at 100% power
        """
        if power < 0:
            # Reverse
            command = COMMAND_SET_B_REV
            pwm = -int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX
        else:
            # Forward / stopped
            command = COMMAND_SET_B_FWD
            pwm = int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX

        # noinspection PyBroadException
        try:
            self.RawWrite(command, [pwm])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motor 2 drive level!')

    def GetMotor2(self):
        """
power = GetMotor2()

Gets the drive level for motor 2, from +1 to -1.
e.g.
0     -> motor 2 is stopped
0.75  -> motor 2 moving forward at 75% power
-0.5  -> motor 2 moving reverse at 50% power
1     -> motor 2 moving forward at 100% power
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_B, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading motor 2 drive level!')
            return

        power = float(i2cRecv[2]) / float(PWM_MAX)

        if i2cRecv[1] == COMMAND_VALUE_FWD:
            return power
        elif i2cRecv[1] == COMMAND_VALUE_REV:
            return -power
        else:
            return

    def SetMotor3(self, power):
        """
SetMotor3(power)

Sets the drive level for motor 3, from +1 to -1.
e.g.
SetMotor3(0)     -> motor 3 is stopped
SetMotor3(0.75)  -> motor 3 moving forward at 75% power
SetMotor3(-0.5)  -> motor 3 moving reverse at 50% power
SetMotor3(1)     -> motor 3 moving forward at 100% power
        """
        if power < 0:
            # Reverse
            command = COMMAND_SET_C_REV
            pwm = -int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX
        else:
            # Forward / stopped
            command = COMMAND_SET_C_FWD
            pwm = int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX

        # noinspection PyBroadException
        try:
            self.RawWrite(command, [pwm])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motor 3 drive level!')

    def GetMotor3(self):
        """
power = GetMotor3()

Gets the drive level for motor 3, from +1 to -1.
e.g.
0     -> motor 3 is stopped
0.75  -> motor 3 moving forward at 75% power
-0.5  -> motor 3 moving reverse at 50% power
1     -> motor 3 moving forward at 100% power
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_C, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading motor 3 drive level!')
            return

        power = float(i2cRecv[2]) / float(PWM_MAX)

        if i2cRecv[1] == COMMAND_VALUE_FWD:
            return power
        elif i2cRecv[1] == COMMAND_VALUE_REV:
            return -power
        else:
            return

    def SetMotor4(self, power):
        """
SetMotor4(power)

Sets the drive level for motor 4, from +1 to -1.
e.g.
SetMotor4(0)     -> motor 4 is stopped
SetMotor4(0.75)  -> motor 4 moving forward at 75% power
SetMotor4(-0.5)  -> motor 4 moving reverse at 50% power
SetMotor4(1)     -> motor 4 moving forward at 100% power
        """
        if power < 0:
            # Reverse
            command = COMMAND_SET_D_REV
            pwm = -int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX
        else:
            # Forward / stopped
            command = COMMAND_SET_D_FWD
            pwm = int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX

        # noinspection PyBroadException
        try:
            self.RawWrite(command, [pwm])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motor 4 drive level!')

    def GetMotor4(self):
        """
power = GetMotor4()

Gets the drive level for motor 4, from +1 to -1.
e.g.
0     -> motor 4 is stopped
0.75  -> motor 4 moving forward at 75% power
-0.5  -> motor 4 moving reverse at 50% power
1     -> motor 4 moving forward at 100% power
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_D, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading motor 4 drive level!')
            return

        power = float(i2cRecv[2]) / float(PWM_MAX)

        if i2cRecv[1] == COMMAND_VALUE_FWD:
            return power
        elif i2cRecv[1] == COMMAND_VALUE_REV:
            return -power
        else:
            return

    def SetMotors(self, power):
        """
SetMotors(power)

Sets the drive level for all motors, from +1 to -1.
e.g.
SetMotors(0)     -> all motors are stopped
SetMotors(0.75)  -> all motors are moving forward at 75% power
SetMotors(-0.5)  -> all motors are moving reverse at 50% power
SetMotors(1)     -> all motors are moving forward at 100% power
        """
        if power < 0:
            # Reverse
            command = COMMAND_SET_ALL_REV
            pwm = -int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX
        else:
            # Forward / stopped
            command = COMMAND_SET_ALL_FWD
            pwm = int(PWM_MAX * power)
            if pwm > PWM_MAX:
                pwm = PWM_MAX

        # noinspection PyBroadException
        try:
            self.RawWrite(command, [pwm])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending all motors drive level!')

    def MotorsOff(self):
        """
MotorsOff()

Sets all motors to stopped, useful when ending a program
        """
        # noinspection PyBroadException
        try:
            self.RawWrite(COMMAND_ALL_OFF, [0])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motors off command!')

    def SetLed(self, state):
        """
SetLed(state)

Sets the current state of the LED, False for off, True for on
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        try:
            self.RawWrite(COMMAND_SET_LED, [level])
        except KeyboardInterrupt:
            raise
        except Exception as e:
            self.Print('Failed sending LED state!')
            self.Print(e)

    def GetLed(self):
        """
state = GetLed()

Reads the current state of the LED, False for off, True for on
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_LED, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading LED state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def ResetEpo(self):
        """
ResetEpo()

Resets the EPO latch state, use to allow movement again after the EPO has been tripped
        """
        # noinspection PyBroadException
        try:
            self.RawWrite(COMMAND_RESET_EPO, [0])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed resetting EPO!')

    def GetEpo(self):
        """
state = GetEpo()

Reads the system EPO latch state.
If False the EPO has not been tripped, and movement is allowed.
If True the EPO has been tripped, movement is disabled if the EPO is not ignored (see SetEpoIgnore)
    Movement can be re-enabled by calling ResetEpo.
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_EPO, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading EPO ignore state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def SetEpoIgnore(self, state):
        """
SetEpoIgnore(state)

Sets the system to ignore or use the EPO latch, set to False if you have an EPO switch, True if you do not
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        # noinspection PyBroadException
        try:
            self.RawWrite(COMMAND_SET_EPO_IGNORE, [level])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending EPO ignore state!')

    def GetEpoIgnore(self):
        """
state = GetEpoIgnore()

Reads the system EPO ignore state, False for using the EPO latch, True for ignoring the EPO latch
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_EPO_IGNORE, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading EPO ignore state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def HasNewIrMessage(self):
        """
state = HasNewIrMessage()

Reads the new IR message received flag.
If False there has been no messages to the IR sensor since the last read.
If True there has been a new IR message which can be read using GetIrMessage().
        """
        try:
            i2cRecv = self.RawRead(COMMAND_GET_NEW_IR, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            self.Print('Failed reading new IR message received flag!')
            raise e

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def GetIrMessage(self):
        """
message = GetIrMessage()

Reads the last IR message which has been received and clears the new IR message received flag.
Returns the bytes from the remote control as a hexadecimal string, e.g. 'F75AD5AA8000'
Use HasNewIrMessage() to see if there has been a new IR message since the last call.
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_LAST_IR, I2C_LONG_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading IR message!')
            return

        message = ''
        for i in range(IR_MAX_BYTES):
            message += '%02X' % (i2cRecv[1 + i])
        return message.rstrip('0')

    def SetLedIr(self, state):
        """
SetLedIr(state)

Sets if IR messages control the state of the LED, False for no effect, True for incoming messages blink the LED
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        # noinspection PyBroadException
        try:
            self.RawWrite(COMMAND_SET_LED_IR, [level])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending LED state!')

    def GetLedIr(self):
        """
state = GetLedIr()

Reads if IR messages control the state of the LED, False for no effect, True for incoming messages blink the LED
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_LED_IR, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading LED state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def GetAnalog1(self):
        """
voltage = GetAnalog1()

Reads the current analog level from port #1 (pin 2).
Returns the value as a voltage based on the 3.3 V reference pin (pin 1).
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_ANALOG_1, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading analog level #1!')
            return

        raw = (i2cRecv[1] << 8) + i2cRecv[2]
        level = float(raw) / float(COMMAND_ANALOG_MAX)
        return level * 3.3

    def GetAnalog2(self):
        """
voltage = GetAnalog2()

Reads the current analog level from port #2 (pin 4).
Returns the value as a voltage based on the 3.3 V reference pin (pin 1).
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_ANALOG_2, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading analog level #2!')
            return

        raw = (i2cRecv[1] << 8) + i2cRecv[2]
        level = float(raw) / float(COMMAND_ANALOG_MAX)
        return level * 3.3

    def SetCommsFailsafe(self, state):
        """
SetCommsFailsafe(state)

Sets the system to enable or disable the communications failsafe
The failsafe will turn the motors off unless it is commanded at least once every 1/4 of a second
Set to True to enable this failsafe, set to False to disable this failsafe
The failsafe is disabled at power on
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        # noinspection PyBroadException
        try:
            self.RawWrite(COMMAND_SET_FAILSAFE, [level])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending communications failsafe state!')

    def GetCommsFailsafe(self):
        """
state = GetCommsFailsafe()

Read the current system state of the communications failsafe, True for enabled, False for disabled
The failsafe will turn the motors off unless it is commanded at least once every 1/4 of a second
        """
        # noinspection PyBroadException
        try:
            i2cRecv = self.RawRead(COMMAND_GET_FAILSAFE, I2C_NORM_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading communications failsafe state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def Help(self):
        """
Help()

Displays the names and descriptions of the various functions and settings provided
        """
        funcList = [ZeroBorg.__dict__.get(a) for a in dir(ZeroBorg) if
                    isinstance(ZeroBorg.__dict__.get(a), types.FunctionType)]
        funcListSorted = sorted(funcList, key=lambda x: x.func_code.co_firstlineno)

        print(self.__doc__)
        print()
        for func in funcListSorted:
            print('=== %s === %s' % (func.func_name, func.func_doc))



import enum
import os
import re
import select
import subprocess
import threading
import time

from platypush.config import Config
from platypush.context import get_bus, get_plugin
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.torrent import TorrentDownloadStartEvent, \
    TorrentDownloadCompletedEvent, TorrentDownloadedMetadataEvent

from platypush.plugins import action
from platypush.utils import find_bins_in_path, find_files_by_ext, \
    is_process_alive, get_ip_or_hostname


class TorrentState(enum.IntEnum):
    IDLE = 1
    DOWNLOADING_METADATA = 2
    DOWNLOADING = 3
    DOWNLOADED = 4


class MediaWebtorrentPlugin(MediaPlugin):
    """
    Plugin to download and stream videos using webtorrent

    Requires:

        * **webtorrent** installed on your system (``npm install -g webtorrent``)
        * **webtorrent-cli** installed on your system (``npm install -g webtorrent-cli``)
        * A media plugin configured for streaming (e.g. media.mplayer, media.vlc, media.mpv or media.omxplayer)
    """

    _supported_media_plugins = {'media.mplayer', 'media.omxplayer', 'media.mpv',
                                'media.vlc', 'media.webtorrent'}

    # Download at least 15 MBs before starting streaming
    _download_size_before_streaming = 15 * 2**20

    _web_stream_ready_timeout = 120

    def __init__(self, webtorrent_bin=None, webtorrent_port=None, *args,
                 **kwargs):
        """
        media.webtorrent will use the default media player plugin you have
        configured (e.g. mplayer, omxplayer, mpv) to stream the torrent.

        :param webtorrent_bin: Path to your webtorrent executable. If not set,
            then Platypush will search for the right executable in your PATH
        :type webtorrent_bin: str

        :param webtorrent_port: Port where the webtorrent will be running
            streaming server will be running (default: 8000)
        :type webtorrent_port: int
        """

        super().__init__(*args, **kwargs)

        self.webtorrent_port = webtorrent_port
        self._webtorrent_process = None
        self._init_webtorrent_bin(webtorrent_bin=webtorrent_bin)
        self._init_media_player()
        self._download_started_event = threading.Event()
        self._torrent_stream_urls = {}

    def _init_webtorrent_bin(self, webtorrent_bin=None):
        if not webtorrent_bin:
            bin_name = 'webtorrent.exe' if os.name == 'nt' else 'webtorrent'
            bins = find_bins_in_path(bin_name)

            if not bins:
                raise RuntimeError('Webtorrent executable not specified and ' +
                                   'not found in your PATH. Make sure that ' +
                                   'webtorrent is either installed or ' +
                                   'configured and that both webtorrent and ' +
                                   'webtorrent-cli are installed')

            self.webtorrent_bin = bins[0]
        else:
            webtorrent_bin = os.path.expanduser(webtorrent_bin)
            if not (os.path.isfile(webtorrent_bin)
                    and (os.name == 'nt' or os.access(webtorrent_bin, os.X_OK))):
                raise RuntimeError('{} is does not exist or is not a valid ' +
                                   'executable file'.format(webtorrent_bin))

            self.webtorrent_bin = webtorrent_bin

    def _init_media_player(self):
        self._media_plugin = None

        for plugin_name in self._supported_media_plugins:
            try:
                if Config.get(plugin_name):
                    self._media_plugin = get_plugin(plugin_name)
                    break
            except:
                pass

        if not self._media_plugin:
            raise RuntimeError(('No media player specified and no ' +
                                'compatible media plugin configured - ' +
                                'supported media plugins: {}').format(
                                    self._supported_media_plugins))

    def _read_process_line(self):
        line = self._webtorrent_process.stdout.readline().decode().strip()
        # Strip output of the colors
        return re.sub('\x1b\[(([0-9]+m)|(.{1,2}))', '', line).strip()

    def _process_monitor(self, resource, download_dir, download_only,
                         player_type, player_args):
        def _thread():
            if not self._webtorrent_process:
                return

            ######
            state = TorrentState.IDLE
            bus = get_bus()
            webtorrent_url = None
            output_dir = None
            media_file = None

            poll = select.poll()
            poll.register(self._webtorrent_process.stdout, select.POLLIN)

            # First wait for the metadata to be ready and the streaming started
            while True:
                result = poll.poll(0)
                if not result:
                    continue

                if not self._is_process_alive():
                    break

                line = self._read_process_line()

                if 'fetching torrent metadata from' in line.lower() \
                        and state == TorrentState.IDLE:
                    # IDLE -> DOWNLOADING_METADATA
                    state = TorrentState.DOWNLOADING_METADATA
                    bus.post(TorrentDownloadedMetadataEvent(resource=resource))
                elif 'downloading: ' in line.lower() \
                        and media_file is None:
                    # Find video files in torrent directory
                    output_dir = os.path.join(
                        download_dir, re.search(
                            'downloading: (.+?)$', line, flags=re.IGNORECASE
                        ).group(1))

                elif 'server running at: ' in line.lower() \
                        and webtorrent_url is None:
                    # Streaming started
                    webtorrent_url = re.search('server running at: (.+?)$',
                                               line, flags=re.IGNORECASE).group(1)
                    webtorrent_url = webtorrent_url.replace(
                        'http://localhost', 'http://' + get_ip_or_hostname())

                    self._torrent_stream_urls[resource] = webtorrent_url
                    self._download_started_event.set()
                    self.logger.info('Torrent stream started on {}'.format(
                        webtorrent_url))

                if output_dir and not media_file:
                    media_files = sorted(find_files_by_ext(
                        output_dir, *self._media_plugin.video_extensions))

                    if media_files:
                        # TODO support for queueing multiple media
                        media_file = os.path.join(output_dir, media_files[0])
                    else:
                        time.sleep(1)  # Wait before the media file is created

                if state.value <= TorrentState.DOWNLOADING_METADATA.value \
                        and media_file and webtorrent_url:
                    # DOWNLOADING_METADATA -> DOWNLOADING
                    state = TorrentState.DOWNLOADING
                    bus.post(TorrentDownloadStartEvent(
                        resource=resource, media_file=media_file,
                        stream_url=webtorrent_url))
                    break

            if not output_dir:
                raise RuntimeError('Could not download torrent')
            if not download_only and (not media_file or not webtorrent_url):
                if not media_file:
                    self.logger.warning(
                        'The torrent does not contain any video files')
                else:
                    self.logger.warning('WebTorrent could not start streaming')

                # Keep downloading but don't start the player
                try: self._webtorrent_process.wait()
                except: pass
                return

            player = None

            if not download_only:
                # Wait until we have enough chunks to start the player
                while True:
                    result = poll.poll(0)
                    if not result:
                        continue

                    if not self._is_process_alive():
                        break

                    try:
                        if os.path.getsize(media_file) > \
                                self._download_size_before_streaming:
                            break
                    except FileNotFoundError:
                        continue

                player = get_plugin('media.' + player_type) if player_type \
                    else self._media_plugin

                media = media_file if player.is_local() else webtorrent_url

                self.logger.info(
                    'Starting playback of {} to {} through {}'.format(
                        media_file, player.__class__.__name__,
                        webtorrent_url))

                subfile = self.get_subtitles(media)
                if subfile:
                    player_args['subtitles'] = subfile

                player.play(media, **player_args)
                self.logger.info('Waiting for player to terminate')

            self._wait_for_player(player)
            self.logger.info('Torrent player terminated')
            bus.post(TorrentDownloadCompletedEvent(resource=resource,
                                                   output_dir=output_dir,
                                                   media_file=media_file))

            try: self.quit()
            except: pass
            self.logger.info('WebTorrent process terminated')

        return _thread

    def _wait_for_player(self, player):
        stop_evt = None

        if player:
            media_cls = player.__class__.__name__

            if media_cls == 'MediaMplayerPlugin':
                stop_evt = player._mplayer_stopped_event
            elif media_cls == 'MediaMpvPlugin' or media_cls == 'MediaVlcPlugin':
                stop_evt = threading.Event()

                def stop_callback():
                    stop_evt.set()
                player.on_stop(stop_callback)
            elif media_cls == 'MediaOmxplayerPlugin':
                stop_evt = threading.Event()

                def stop_callback():
                    stop_evt.set()
                player.add_handler('stop', stop_callback)

        if stop_evt:
            stop_evt.wait()
        else:
            # Fallback: wait for the webtorrent process to terminate
            self._webtorrent_process.wait()

    def _get_torrent_download_dir(self):
        if self._media_plugin.download_dir:
            return self._media_plugin.download_dir
        else:
            d = os.path.join(os.path.expanduser('~'), 'Downloads')
            os.makedirs(d, exist_ok=True)
            return d

    def get_subtitles(self, filepath):
        try:
            plugin = get_plugin('media.subtitles')
            if not plugin or not plugin.languages:
                return

            subs = plugin.get_subtitles(filepath).output
            if not subs:
                return

            sub = plugin.download_subtitles(subs[0]['SubDownloadLink'],
                                            filepath).output

            if sub:
                return sub['filename']
        except Exception as e:
            self.logger.warning('Could not get subtitles for {}: {}'.format(
                filepath, str(e)))

    @action
    def play(self, resource, player=None, download_only=False, **player_args):
        """
        Download and stream a torrent

        :param resource: Play a resource, as a magnet link, torrent URL or
            torrent file path
        :type resource: str

        :param player: If set, use this plugin type as a player for the
            torrent. Supported types: 'mplayer', 'vlc', 'omxplayer', 'chromecast', 'mpv'.
            If not set, then the default configured media plugin will be used.
        :type player: str

        :param player_args: Any arguments to pass to the player plugin's
            play() method
        :type player_args: dict

        :param download_only: If false then it will start streaming the torrent on the local player once the
            download starts, otherwise it will just download it (default: false)
        :type download_only: bool
        """

        if self._webtorrent_process:
            try:
                self.quit()
            except Exception as e:
                self.logger.debug('Failed to quit the previous instance: {}'.
                                  format(str(e)))

        download_dir = self._get_torrent_download_dir()
        webtorrent_args = [self.webtorrent_bin, 'download', '-o', download_dir]

        if self.webtorrent_port:
            webtorrent_args += ['-p', self.webtorrent_port]
        webtorrent_args += [resource]

        self._download_started_event.clear()
        self._webtorrent_process = subprocess.Popen(webtorrent_args,
                                                    stdout=subprocess.PIPE)

        threading.Thread(target=self._process_monitor(
            resource=resource, download_dir=download_dir,
            player_type=player, player_args=player_args,
            download_only=download_only)).start()

        stream_url = None
        player_ready_wait_start = time.time()

        while not stream_url:
            triggered = self._download_started_event.wait(
                self._web_stream_ready_timeout)

            if not triggered or time.time() - player_ready_wait_start >= \
                    self._web_stream_ready_timeout:
                break

            stream_url = self._torrent_stream_urls.get(resource)

        if not stream_url:
            return (None, ('The webtorrent process hasn\'t started ' +
                           'streaming after {} seconds').format(
                self._web_stream_ready_timeout))

        return {'resource': resource, 'url': stream_url}

    @action
    def download(self, resource, **kwargs):
        return self.play(resource, download_only=True)

    @action
    def stop(self):
        """ Stop the playback """
        return self.quit()

    @action
    def quit(self):
        """ Quit the player """
        if self._is_process_alive():
            self._webtorrent_process.terminate()
            self._webtorrent_process.wait()
            try: self._webtorrent_process.kill()
            except: pass

        self._webtorrent_process = None

    @action
    def load(self, resource, **kwargs):
        """
        Load a torrent resource in the player.
        """
        return self.play(resource)

    def _is_process_alive(self):
        return is_process_alive(self._webtorrent_process.pid) \
            if self._webtorrent_process else False

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "state": "play"  # or "stop" or "pause"
            }
        """

        return {'state': self._media_plugin.status().get('state', PlayerState.STOP.value)}


# vim:sw=4:ts=4:et:



import json
import os
import requests

from platypush.context import get_backend
from platypush.plugins import Plugin, action


class PushbulletPlugin(Plugin):
    """
    This plugin allows you to send pushes and files to your PushBullet account.
    Note: This plugin will only work if the :mod:`platypush.backend.pushbullet`
    backend is configured.

    Requires:

        * **requests** (``pip install requests``)
        * The :class:`platypush.backend.pushbullet.Pushbullet` backend enabled
    """

    def __init__(self, token: str = None, *args, **kwargs):
        """
        :param token: Pushbullet API token. If not set the plugin will try to retrieve it from
            the Pushbullet backend configuration, if available
        """
        super().__init__(*args, **kwargs)

        if not token:
            backend = get_backend('pushbullet')
            if not backend or not backend.token:
                raise AttributeError('No Pushbullet token specified')

            self.token = backend.token
        else:
            self.token = token

        self._devices = []
        self._devices_by_id = {}
        self._devices_by_name = {}

    @action
    def get_devices(self):
        """
        Get the list of available devices
        """
        resp = requests.get('https://api.pushbullet.com/v2/devices',
                headers={'Authorization': 'Bearer ' + self.token,
                    'Content-Type': 'application/json'})

        self._devices = resp.json().get('devices', [])
        self._devices_by_id = {
                dev['iden']: dev
                for dev in self._devices
        }

        self._devices_by_name = {
                dev['nickname']: dev
                for dev in self._devices if 'nickname' in dev
        }

    @action
    def get_device(self, device):
        """
        :param device: Device ID or name
        """
        output = None
        refreshed = False

        while not output:
            if device in self._devices_by_id:
                return self._devices_by_id[device]
            if device in self._devices_by_name:
                return self._devices_by_name[device]
            if refreshed:
                return None

            self.get_devices()
            refreshed = True

    @action
    def send_note(self, device: str = None, body: str = None, title: str = None, url: str = None, **kwargs):
        """
        Send a note push.

        :param device: Device ID or name (default: None, all devices)
        :param body: Note body
        :param title: Note title
        :param url: URL attached to the note
        :param kwargs: Push arguments, see https://docs.pushbullet.com/#create-push
        """

        if device:
            device = self.get_device(device).output
            if not device:
                raise RuntimeError('No such device')

        kwargs['body'] = body
        kwargs['title'] = title
        kwargs['type'] = 'link' if url else 'note'

        if device:
            kwargs['device_iden'] = device['iden']

        resp = requests.post('https://api.pushbullet.com/v2/pushes',
                             data=json.dumps(kwargs),
                             headers={'Authorization': 'Bearer ' + self.token,
                                      'Content-Type': 'application/json'})

        if resp.status_code >= 400:
            raise Exception('Pushbullet push failed with status {}: {}'.
                            format(resp.status_code, resp.json()))

    @action
    def send_file(self, filename: str, device: str = None):
        """
        Send a file.

        :param device: Device ID or name (default: None, all devices)
        :param filename: Path to the local file
        """

        if device:
            device = self.get_device(device).output
            if not device:
                raise RuntimeError('No such device')

        pushbullet = get_backend('pushbullet')
        resp = requests.post('https://api.pushbullet.com/v2/upload-request',
                             data=json.dumps({'file_name': os.path.basename(filename)}),
                             headers={'Authorization': 'Bearer ' + self.token,
                                      'Content-Type': 'application/json'})

        if resp.status_code != 200:
            raise Exception('Pushbullet file upload request failed with status {}'.
                            format(resp.status_code))

        r = resp.json()
        resp = requests.post(r['upload_url'], data=r['data'],
                             files={'file': open(filename, 'rb')})

        if resp.status_code != 204:
            raise Exception('Pushbullet file upload failed with status {}'.
                            format(resp.status_code))

        resp = requests.post('https://api.pushbullet.com/v2/pushes',
                             headers={'Authorization': 'Bearer ' + self.token,
                                      'Content-Type': 'application/json'},

                             data=json.dumps({
                                 'type': 'file',
                                 'device_iden': device['iden'] if device else None,
                                 'file_name': r['file_name'],
                                 'file_type': r['file_type'],
                                 'file_url': r['file_url'] }))

        if resp.status_code >= 400:
            raise Exception('Pushbullet file push failed with status {}'.
                            format(resp.status_code))

        return {
            'filename': r['file_name'],
            'type': r['file_type'],
            'url': r['file_url']
        }


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class NewWeatherConditionEvent(Event):
    """
    Event triggered when the weather condition changes
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NewPrecipitationForecastEvent(Event):
    """
    Event triggered when the precipitation forecast changes
    """
    def __init__(self, *args, average: float, total: float, time_frame: int, **kwargs):
        super().__init__(*args, average=average, total=total, time_frame=time_frame, **kwargs)


# vim:sw=4:ts=4:et:



import io
import json
import os
import threading

from typing import Any, Optional, IO

from platypush.message import Message
from platypush.plugins import Plugin, action


class MqttPlugin(Plugin):
    """
    This plugin allows you to send custom message to a message queue compatible
    with the MQTT protocol, see http://mqtt.org/

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)

    """

    def __init__(self, host=None, port=1883, tls_cafile=None,
                 tls_certfile=None, tls_keyfile=None,
                 tls_version=None, tls_ciphers=None, username=None,
                 password=None, **kwargs):
        """
        :param host: If set, MQTT messages will by default routed to this host unless overridden in `send_message` (default: None)
        :type host: str

        :param port: If a default host is set, specify the listen port (default: 1883)
        :type port: int

        :param tls_cafile: If a default host is set and requires TLS/SSL, specify the certificate authority file (default: None)
        :type tls_cafile: str

        :param tls_certfile: If a default host is set and requires TLS/SSL, specify the certificate file (default: None)
        :type tls_certfile: str

        :param tls_keyfile: If a default host is set and requires TLS/SSL, specify the key file (default: None)
        :type tls_keyfile: str

        :param tls_version: If a default host is set and requires TLS/SSL, specify the minimum TLS supported version (default: None)
        :type tls_version: str

        :param tls_ciphers: If a default host is set and requires TLS/SSL, specify the supported ciphers (default: None)
        :type tls_ciphers: str

        :param username: If a default host is set and requires user authentication, specify the username ciphers (default: None)
        :type username: str

        :param password: If a default host is set and requires user authentication, specify the password ciphers (default: None)
        :type password: str
        """

        super().__init__(**kwargs)

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.tls_cafile = os.path.abspath(os.path.expanduser(tls_cafile)) \
            if tls_cafile else None

        self.tls_certfile = os.path.abspath(os.path.expanduser(tls_certfile)) \
            if tls_certfile else None

        self.tls_keyfile = os.path.abspath(os.path.expanduser(tls_keyfile)) \
            if tls_keyfile else None

        self.tls_version = tls_version
        self.tls_ciphers = tls_ciphers

    @action
    def publish(self, topic: str, msg: Any, host: Optional[str] = None, port: int = 1883,
                reply_topic: Optional[str] = None, timeout: int = 60,
                tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                tls_keyfile: Optional[str] = None, tls_version: Optional[str] = None,
                tls_ciphers: Optional[str] = None, username: Optional[str] = None,
                password: Optional[str] = None):
        """
        Sends a message to a topic.

        :param topic: Topic/channel where the message will be delivered
        :param msg: Message to be sent. It can be a list, a dict, or a Message object.
        :param host: MQTT broker hostname/IP.
        :param port: MQTT broker port (default: 1883).
        :param reply_topic: If a ``reply_topic`` is specified, then the action will wait for a response on this topic.
        :param timeout: If ``reply_topic`` is set, use this parameter to specify the maximum amount of time to
            wait for a response (default: 60 seconds).
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the certificate requires a certificate authority
            to authenticate it, `ssl_cafile` will point to the provided ca.crt file (default: None).
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a client certificate it required, specify it
            here (default: None).
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a client certificate key it required, specify
            it here (default: None).
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it requires a certain TLS version, specify it
            here (default: None).
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an explicit list of supported ciphers is
            required, specify it here (default: None).
        :param username: Specify it if the MQTT server requires authentication (default: None).
        :param password: Specify it if the MQTT server requires authentication (default: None).
        """
        from paho.mqtt.client import Client

        if not host and not self.host:
            raise RuntimeError('No host specified and no default host configured')

        if not host:
            tls_cafile = self.tls_cafile
            tls_certfile = self.tls_certfile
            tls_keyfile = self.tls_keyfile
            tls_version = self.tls_version
            tls_ciphers = self.tls_ciphers
            username = self.username
            password = self.password

        client = Client()

        if username and password:
            client.username_pw_set(username, password)
        if tls_cafile:
            client.tls_set(ca_certs=tls_cafile, certfile=tls_certfile, keyfile=tls_keyfile, tls_version=tls_version,
                           ciphers=tls_ciphers)

        # Try to parse it as a platypush message or dump it to JSON from a dict/list
        if isinstance(msg, dict) or isinstance(msg, list):
            msg = json.dumps(msg)

            # noinspection PyBroadException
            try:
                msg = Message.build(json.loads(msg))
            except:
                pass

        client.connect(host, port, keepalive=timeout)
        response_buffer = io.BytesIO()

        try:
            response_received = threading.Event()

            if reply_topic:
                client.on_message = self._response_callback(reply_topic=reply_topic,
                                                            event=response_received,
                                                            buffer=response_buffer)
                client.subscribe(reply_topic)

            client.publish(topic, str(msg))
            if not reply_topic:
                return

            client.loop_start()
            ok = response_received.wait(timeout=timeout)
            if not ok:
                raise TimeoutError('Response timed out')
            return response_buffer.getvalue()
        finally:
            response_buffer.close()

            # noinspection PyBroadException
            try:
                client.loop_stop()
            except:
                pass

            client.disconnect()

    @staticmethod
    def _response_callback(reply_topic: str, event: threading.Event, buffer: IO[bytes]):
        def on_message(client, _, msg):
            if msg.topic != reply_topic:
                return

            buffer.write(msg.payload)
            client.loop_stop()
            event.set()
        return on_message

    @action
    def send_message(self, *args, **kwargs):
        """
        Alias for :meth:`platypush.plugins.mqtt.MqttPlugin.publish`.
        """
        return self.publish(*args, **kwargs)


# vim:sw=4:ts=4:et:



import base64
import os

from platypush.plugins import Plugin, action


class DropboxPlugin(Plugin):
    """
    Plugin to manage a Dropbox account and its files and folders.

    Requires:

        * **dropbox** (``pip install dropbox``)
    """

    def __init__(self, access_token, **kwargs):
        """
        :param access_token: Dropbox API access token. You can get yours by creating an app on
            https://dropbox.com/developers/apps
        :type access_token: str
        """

        super().__init__(**kwargs)
        self.access_token = access_token
        self._dbx = None

    def _get_instance(self):
        """
        :rtype: :class:`dropbox.Dropbox`
        """
        # noinspection PyPackageRequirements
        import dropbox

        if not self._dbx:
            self._dbx = dropbox.Dropbox(self.access_token)

        return self._dbx

    @action
    def get_account(self, account_id=None):
        """
        Get the information about a linked Dropbox account

        :param account_id: account_id. If none is specified then it will retrieve the current user's account id
        :type account_id: str

        :return: dict with the following attributes:
            account_id, name, email, email_verified, disabled, profile_photo_url, team_member_id
        """

        dbx = self._get_instance()

        if not account_id:
            acc = dbx.users_get_current_account()
        else:
            acc = dbx.users_get_account(account_id)

        return {
            'account_id': acc.account_id,
            'name': acc.name.display_name,
            'email': acc.email,
            'email_verified': acc.email_verified,
            'disabled': acc.disabled,
            'profile_photo_url': acc.profile_photo_url,
            'team_member_id': acc.team_member_id,
        }

    @action
    def get_usage(self):
        """
        Get the amount of allocated and used remote space

        :return: dict
        """

        dbx = self._get_instance()
        usage = dbx.users_get_space_usage()

        return {
            'used': usage.used,
            'allocated': usage.allocation.get_individual().allocated,
        }

    @action
    def list(self, folder=''):
        """
        Returns the files in a folder

        :param folder: Folder name (default: root)
        :type folder: str

        :return: dict
        """

        from dropbox.files import FileMetadata

        dbx = self._get_instance()
        files = dbx.files_list_folder(folder).entries
        entries = []

        for item in files:
            entry = {
                attr: getattr(item, attr)
                for attr in ['id', 'name', 'path_display', 'path_lower',
                             'parent_shared_folder_id', 'property_groups']
            }

            if item.sharing_info:
                entry['sharing_info'] = {
                    attr: getattr(item.sharing_info, attr)
                    for attr in ['no_access', 'parent_shared_folder_id', 'read_only',
                                 'shared_folder_id', 'traverse_only']
                }
            else:
                entry['sharing_info'] = {}

            if isinstance(item, FileMetadata):
                entry['client_modified'] = item.client_modified.isoformat()
                entry['server_modified'] = item.server_modified.isoformat()

                for attr in ['content_hash', 'has_explicit_shared_members', 'is_downloadable', 'rev', 'size']:
                    if hasattr(item, attr):
                        entry[attr] = getattr(item, attr)

            entries.append(entry)

        return entries

    @action
    def copy(self, from_path: str, to_path: str, allow_shared_folder=True, autorename=False,
             allow_ownership_transfer=False):
        """
        Copy a file or folder to a different location in the user's Dropbox. If the source path is a folder all
        its contents will be copied.

        :param from_path: Source path
        :param to_path: Destination path
        :param bool allow_shared_folder: If true, :meth:`files_copy` will copy
            contents in shared folder, otherwise
            ``RelocationError.cant_copy_shared_folder`` will be returned if
            ``from_path`` contains shared folder. This field is always true for
            :meth:`files_move`.
        :param bool autorename: If there's a conflict, have the Dropbox server
            try to autorename the file to avoid the conflict.
        :param bool allow_ownership_transfer: Allow moves by owner even if it
            would result in an ownership transfer for the content being moved.
            This does not apply to copies.
        """

        dbx = self._get_instance()
        dbx.files_copy_v2(from_path, to_path, allow_shared_folder=allow_shared_folder,
                          autorename=autorename, allow_ownership_transfer=allow_ownership_transfer)

    @action
    def move(self, from_path: str, to_path: str, allow_shared_folder=True, autorename=False,
             allow_ownership_transfer=False):
        """
        Move a file or folder to a different location in the user's Dropbox. If the source path is a folder all its
        contents will be moved.

        :param from_path: Source path
        :param to_path: Destination path
        :param bool allow_shared_folder: If true, :meth:`files_copy` will copy
            contents in shared folder, otherwise
            ``RelocationError.cant_copy_shared_folder`` will be returned if
            ``from_path`` contains shared folder. This field is always true for
            :meth:`files_move`.
        :param bool autorename: If there's a conflict, have the Dropbox server
            try to autorename the file to avoid the conflict.
        :param bool allow_ownership_transfer: Allow moves by owner even if it
            would result in an ownership transfer for the content being moved.
            This does not apply to copies.
        """

        dbx = self._get_instance()
        dbx.files_move_v2(from_path, to_path, allow_shared_folder=allow_shared_folder,
                          autorename=autorename, allow_ownership_transfer=allow_ownership_transfer)

    @action
    def delete(self, path: str):
        """
        Delete the file or folder at a given path. If the path is a folder, all its contents will be deleted too

        :param str path: Path to be removed
        """

        dbx = self._get_instance()
        dbx.files_delete_v2(path)

    @action
    def restore(self, path: str, rev: str):
        """
        Restore a specific revision of a file to the given path.

        :param str path: Path to be removed
        :param str rev: Revision ID to be restored
        """

        dbx = self._get_instance()
        dbx.files_restore(path, rev)

    @action
    def mkdir(self, path: str):
        """
        Create a folder at a given path.

        :param str path: Folder path
        """

        dbx = self._get_instance()
        dbx.files_create_folder_v2(path)

    @action
    def save(self, path: str, url: str):
        """
        Save a specified URL into a file in user's Dropbox. If the given path already exists, the file will be renamed
        to avoid the conflict (e.g. myfile (1).txt).

        :param path: Dropbox destination path
        :param url: URL to download
        """

        dbx = self._get_instance()
        dbx.files_save_url(path, url)

    def _file_download(self, path: str, download_path=None):
        dbx = self._get_instance()
        metadata, response = dbx.files_download(path)
        ret = self._parse_metadata(metadata)
        ret['encoding'] = response.encoding or response.apparent_encoding

        if download_path:
            if os.path.isdir(download_path):
                download_path = os.path.join(download_path, metadata.name)

            with open(download_path, 'wb') as f:
                f.write(response.content)
            ret['file'] = download_path
        else:
            if ret['encoding'] in ('ascii', 'unicode', 'utf-8'):
                ret['content'] = response.text
            else:
                ret['content'] = base64.encodebytes(response.content).decode().strip()

        return ret

    def _file_download_zip(self, path: str, download_path=None):
        dbx = self._get_instance()
        result, response = dbx.files_download_zip(path)
        ret = self._parse_metadata(result.metadata)

        if download_path:
            if os.path.isdir(download_path):
                download_path = os.path.join(download_path, result.metadata.name + '.zip')

            with open(download_path, 'wb') as f:
                f.write(response.content)
            ret['file'] = download_path
        else:
            ret['content'] = base64.encodebytes(response.content).decode().strip()

        return ret

    # noinspection PyShadowingBuiltins
    @action
    def download(self, path: str, download_path=None, zip=False):
        """
        Download a file or a zipped directory from a user's Dropbox.

        :param str path: Dropbox destination path
        :param str download_path: Destination path on the local machine (optional)
        :param bool zip: If set then the content will be downloaded in zip format (default: False)
        :rtype: dict
        :return: A dictionary with keys: ``('id', 'name', 'parent_shared_folder_id', 'path', 'size', 'encoding',
            'content_hash')``. If download_path is set 'file' is also returned. Otherwise 'content' will be returned.
            If it's a text file then 'content' will contain its string representation, otherwise its base64-encoded
            representation.
        """

        from dropbox.files import FolderMetadata

        if download_path:
            download_path = os.path.abspath(os.path.expanduser(download_path))

        dbx = self._get_instance()

        metadata = dbx.files_get_metadata(path)
        if isinstance(metadata, FolderMetadata):
            zip = True

        if zip:
            return self._file_download_zip(path, download_path)

        return self._file_download(path, download_path)

    @action
    def get_metadata(self, path: str):
        """
        Get the metadata of the specified path

        :param str path: Path to the resource
        :rtype dict:
        """

        dbx = self._get_instance()
        metadata = dbx.files_get_metadata(path)
        return self._parse_metadata(metadata)

    @staticmethod
    def _parse_metadata(metadata):
        from dropbox.files import FileMetadata, FolderMetadata

        ret = {
            'name': metadata.name,
            'parent_shared_folder_id': metadata.parent_shared_folder_id,
            'path': metadata.path_display,
        }

        if isinstance(metadata, FileMetadata):
            ret['type'] = 'file'
            ret['id'] = metadata.id
            ret['size'] = metadata.size
            ret['content_hash'] = metadata.content_hash
            ret['rev'] = metadata.rev
        elif isinstance(metadata, FolderMetadata):
            ret['type'] = 'folder'
            ret['id'] = metadata.id
            ret['shared_folder_id'] = metadata.shared_folder_id

        return ret

    @action
    def search(self, query: str, path='', start=0, max_results=100, content=False):
        """
        Searches for files and folders. Note: Recent changes may not immediately
        be reflected in search results due to a short delay in indexing.

        :param str path: The path in the user's Dropbox to search. Should probably be a folder.
        :param str query: The string to search for. The search string is split
            on spaces into multiple tokens. For file name searching, the last
            token is used for prefix matching (i.e. "bat c" matches "bat cave"
            but not "batman car").
        :param long start: The starting index within the search results (used for paging).
        :param long max_results: The maximum number of search results to return.
        :param content: Search also in files content (default: False)
        :rtype dict:
        :return: Dictionary with the following fields: ``('matches', 'start')``.
        """

        from dropbox.files import SearchMode

        dbx = self._get_instance()
        response = dbx.files_search(query=query, path=path, start=start, max_results=max_results,
                                    mode=SearchMode.filename_and_content if content else SearchMode.filename)

        results = [self._parse_metadata(match.metadata) for match in response.matches]

        return {
            'results': results,
            'start': response.start,
        }

    @action
    def upload(self, file=None, text=None, path='/', overwrite=False, autorename=False):
        """
        Create a new file with the contents provided in the request. Do not use this to upload a file larger than 150 MB

        :param str file: File to be uploaded
        :param str text: Text content to be uploaded
        :param str path: Path in the user's Dropbox to save the file.
        :param bool overwrite: If set, in case of conflict the file will be overwritten (default: append content)
        :param bool autorename: If there's a conflict, as determined by
            ``mode``, have the Dropbox server try to autorename the file to
            avoid conflict.
        :rtype dict:
        :return: Dictionary with the metadata of the uploaded file
        """

        from dropbox.files import WriteMode, FolderMetadata
        from dropbox.exceptions import ApiError

        dbx = self._get_instance()

        if file:
            file = os.path.abspath(os.path.expanduser(file))
            try:
                metadata = dbx.files_get_metadata(path)
                if isinstance(metadata, FolderMetadata):
                    path = '/'.join([path, os.path.basename(file)])
            except ApiError:
                pass

            with open(file, 'rb') as f:
                content = f.read()
        elif text:
            content = text.encode()
        else:
            raise SyntaxError('Please specify either a file or text to be uploaded')

        metadata = dbx.files_upload(content, path, autorename=autorename,
                                    mode=WriteMode.overwrite if overwrite else WriteMode.add)

        return self._parse_metadata(metadata)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class IrSensorEvent(Event):
    """
    Base class for infrared sensor events
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class IrKeyUpEvent(IrSensorEvent):
    """
    Event triggered when a key on an infrared remote is released
    """

    def __init__(self, message=None, *args, **kwargs):
        """
        :param message: The received infrared message
        """

        super().__init__(*args, message=message, **kwargs)


class IrKeyDownEvent(IrSensorEvent):
    """
    Event triggered when a key on an infrared remote is pressed
    """

    def __init__(self, message=None, *args, **kwargs):
        """
        :param message: The received infrared message
        """

        super().__init__(*args, message=message, **kwargs)


# vim:sw=4:ts=4:et:




from typing import List

from platypush.message.response.camera import CameraResponse


class AndroidCameraStatusResponse(CameraResponse):
    """
    Sample response::

        .. code-block:: json

        {
            "stream_url": "https://192.168.1.30:8080/video",
            "image_url": "https://192.168.1.30:8080/photo.jpg",
            "audio_url": "https://192.168.1.30:8080/audio.wav",
            "orientation": "landscape",
            "idle": "off",
            "audio_only": "off",
            "overlay": "off",
            "quality": "49",
            "focus_homing": "off",
            "ip_address": "192.168.1.30",
            "motion_limit": "250",
            "adet_limit": "200",
            "night_vision": "off",
            "night_vision_average": "2",
            "night_vision_gain": "1.0",
            "motion_detect": "off",
            "motion_display": "off",
            "video_chunk_len": "60",
            "gps_active": "off",
            "video_size": "1920x1080",
            "mirror_flip": "none",
            "ffc": "off",
            "rtsp_video_formats": "",
            "rtsp_audio_formats": "",
            "video_connections": "0",
            "audio_connections": "0",
            "ivideon_streaming": "off",
            "zoom": "100",
            "crop_x": "50",
            "crop_y": "50",
            "coloreffect": "none",
            "scenemode": "auto",
            "focusmode": "continuous-video",
            "whitebalance": "auto",
            "flashmode": "off",
            "antibanding": "off",
            "torch": "off",
            "focus_distance": "0.0",
            "focal_length": "4.25",
            "aperture": "1.7",
            "filter_density": "0.0",
            "exposure_ns": "9384",
            "frame_duration": "33333333",
            "iso": "100",
            "manual_sensor": "off",
            "photo_size": "1920x1080"
        }

    """

    attrs = {
        'orientation', 'idle', 'audio_only', 'overlay', 'quality', 'focus_homing',
        'ip_address', 'motion_limit', 'adet_limit', 'night_vision',
        'night_vision_average', 'night_vision_gain', 'motion_detect',
        'motion_display', 'video_chunk_len', 'gps_active', 'video_size',
        'mirror_flip', 'ffc', 'rtsp_video_formats', 'rtsp_audio_formats',
        'video_connections', 'audio_connections', 'ivideon_streaming', 'zoom',
        'crop_x', 'crop_y', 'coloreffect', 'scenemode', 'focusmode',
        'whitebalance', 'flashmode', 'antibanding', 'torch', 'focus_distance',
        'focal_length', 'aperture', 'filter_density', 'exposure_ns',
        'frame_duration', 'iso', 'manual_sensor', 'photo_size',
    }

    def __init__(self, *args,
                 name: str = None,
                 stream_url: str = None,
                 image_url: str = None,
                 audio_url: str = None,
                 orientation: str = None,
                 idle: str = None,
                 audio_only: str = None,
                 overlay: str = None,
                 quality: str = None,
                 focus_homing: str = None,
                 ip_address: str = None,
                 motion_limit: str = None,
                 adet_limit: str = None,
                 night_vision: str = None,
                 night_vision_average: str = None,
                 night_vision_gain: str = None,
                 motion_detect: str = None,
                 motion_display: str = None,
                 video_chunk_len: str = None,
                 gps_active: str = None,
                 video_size: str = None,
                 mirror_flip: str = None,
                 ffc: str = None,
                 rtsp_video_formats: str = None,
                 rtsp_audio_formats: str = None,
                 video_connections: str = None,
                 audio_connections: str = None,
                 ivideon_streaming: str = None,
                 zoom: str = None,
                 crop_x: str = None,
                 crop_y: str = None,
                 coloreffect: str = None,
                 scenemode: str = None,
                 focusmode: str = None,
                 whitebalance: str = None,
                 flashmode: str = None,
                 antibanding: str = None,
                 torch: str = None,
                 focus_distance: str = None,
                 focal_length: str = None,
                 aperture: str = None,
                 filter_density: str = None,
                 exposure_ns: str = None,
                 frame_duration: str = None,
                 iso: str = None,
                 manual_sensor: str = None,
                 photo_size: str = None,
                 **kwargs):

        self.status = {
            "name": name,
            "stream_url": stream_url,
            "image_url": image_url,
            "audio_url": audio_url,
            "orientation": orientation,
            "idle": True if idle == "on" else False,
            "audio_only": True if audio_only == "on" else False,
            "overlay": True if overlay == "on" else False,
            "quality": int(quality or 0),
            "focus_homing": True if focus_homing == "on" else False,
            "ip_address": ip_address,
            "motion_limit": int(motion_limit or 0),
            "adet_limit": int(adet_limit or 0),
            "night_vision": True if night_vision == "on" else False,
            "night_vision_average": float(night_vision_average or 0),
            "night_vision_gain": float(night_vision_gain or 0),
            "motion_detect": True if motion_detect == "on" else False,
            "motion_display": True if motion_display == "on" else False,
            "video_chunk_len": int(video_chunk_len or 0),
            "gps_active": True if gps_active == "on" else False,
            "video_size": video_size,
            "mirror_flip": mirror_flip,
            "ffc": True if ffc == "on" else False,
            "rtsp_video_formats": rtsp_video_formats,
            "rtsp_audio_formats": rtsp_audio_formats,
            "video_connections": int(video_connections or 0),
            "audio_connections": int(audio_connections or 0),
            "ivideon_streaming": True if ivideon_streaming == "on" else False,
            "zoom": int(zoom or 0),
            "crop_x": int(crop_x or 0),
            "crop_y": int(crop_y or 0),
            "coloreffect": coloreffect,
            "scenemode": scenemode,
            "focusmode": focusmode,
            "whitebalance": whitebalance,
            "flashmode": True if flashmode == "on" else False,
            "antibanding": True if antibanding == "on" else False,
            "torch": True if torch == "on" else False,
            "focus_distance": float(focus_distance or 0),
            "focal_length": float(focal_length or 0),
            "aperture": float(aperture or 0),
            "filter_density": float(filter_density or 0),
            "exposure_ns": int(exposure_ns or 0),
            "frame_duration": int(frame_duration or 0),
            "iso": int(iso or 0),
            "manual_sensor": True if manual_sensor == "on" else False,
            "photo_size": photo_size,
        }

        super().__init__(*args, output=self.status, **kwargs)


class AndroidCameraStatusListResponse(CameraResponse):
    def __init__(self, status: List[AndroidCameraStatusResponse], **kwargs):
        self.status = [s.status for s in status]
        super().__init__(output=self.status, **kwargs)


class AndroidCameraPictureResponse(CameraResponse):
    def __init__(self, image_file: str,  *args, **kwargs):
        self.image_file = image_file
        super().__init__(*args, output={'image_file': image_file}, **kwargs)


# vim:sw=4:ts=4:et:



import json
import os

from flask import Blueprint, render_template, request

from platypush.backend.http.app import template_folder, static_folder
from platypush.backend.http.app.utils import authenticate, get_websocket_port

from platypush.backend.http.utils import HttpUtils
from platypush.config import Config

index = Blueprint('index', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    index,
]


@index.route('/')
@authenticate()
def index():
    """ Route to the main web panel """
    configured_plugins = Config.get_plugins()
    enabled_templates = {}
    enabled_scripts = {}
    enabled_styles = {}

    enabled_plugins = set(request.args.get('enabled_plugins', '').split(','))
    for plugin in enabled_plugins:
        if plugin not in configured_plugins:
            configured_plugins[plugin] = {}

    configured_plugins['execute'] = {}
    disabled_plugins = set(request.args.get('disabled_plugins', '').split(','))

    js_folder = os.path.abspath(
        os.path.join(template_folder, '..', 'static', 'js'))
    style_folder = os.path.abspath(
        os.path.join(template_folder, '..', 'static', 'css', 'dist'))

    for plugin, conf in configured_plugins.copy().items():
        if plugin in disabled_plugins:
            if plugin == 'execute':
                configured_plugins.pop('execute')
            continue

        template_file = os.path.join(
            template_folder, 'plugins', plugin, 'index.html')

        script_file = os.path.join(js_folder, 'plugins', plugin, 'index.js')
        style_file = os.path.join(style_folder, 'webpanel', 'plugins', plugin+'.css')

        if os.path.isfile(template_file):
            conf['_template_file'] = '/' + '/'.join(template_file.split(os.sep)[-3:])
            enabled_templates[plugin] = conf

        if os.path.isfile(script_file):
            conf['_script_file'] = '/'.join(script_file.split(os.sep)[-4:])
            enabled_scripts[plugin] = conf

        if os.path.isfile(style_file):
            conf['_style_file'] = 'css/dist/' + style_file[len(style_folder)+1:]
            enabled_styles[plugin] = conf

    http_conf = Config.get('backend.http')
    return render_template('index.html', templates=enabled_templates,
                           scripts=enabled_scripts, styles=enabled_styles,
                           utils=HttpUtils, token=Config.get('token'),
                           websocket_port=get_websocket_port(),
                           template_folder=template_folder, static_folder=static_folder,
                           plugins=Config.get_plugins(), backends=Config.get_backends(),
                           procedures=json.dumps(Config.get_procedures()),
                           has_ssl=http_conf.get('ssl_cert') is not None)


# vim:sw=4:ts=4:et:
